@@ -24,13 +24,16 @@ static struct listHead authlist = { &authlist, &authlist };
 
 bool getUserPass(const char *url, char *creds, bool find_proxy)
 {
-	const char *host = getHostURL(url);
 	int port = getPortURL(url);
+	char host[MAXHOSTLEN];
 	const char *dir, *dirend;
 	struct httpAuth *a;
 	struct httpAuth *found = NULL;
 	int d1len, d2len;
 
+	if (!getProtHostURL(url, NULL, host))
+		return false;
+
 	getDirURL(url, &dir, &dirend);
 	d2len = dirend - dir;
 
@@ -58,11 +61,14 @@ bool getUserPass(const char *url, char *creds, bool find_proxy)
 
 bool getUserPassRealm(const char *url, char *creds, const char *realm)
 {
-	const char *host = getHostURL(url);
+	char host[MAXHOSTLEN];
 	int port = getPortURL(url);
 	struct httpAuth *a;
 	struct httpAuth *found = NULL;
 
+	if (!getProtHostURL(url, NULL, host))
+		return false;
+
 	foreach(a, authlist) {
 		if (found == NULL && stringEqualCI(a->host, host) &&
 		    a->port == port) {
@@ -82,15 +88,14 @@ bool getUserPassRealm(const char *url, char *creds, const char *realm)
 
 bool
 addWebAuthorization(const char *url,
-		    const char *credentials, bool proxy,
-		    const char *realm)
+		    const char *credentials, bool proxy, const char *realm)
 {
+	char host[MAXHOSTLEN];
 	struct httpAuth *a;
-	const char *host;
 	const char *dir = 0, *dirend;
 	int port, dl;
 	bool urlProx = isProxyURL(url);
-	bool updated = true;
+	bool updated = false;
 
 	if (proxy) {
 		if (!urlProx) {
@@ -100,7 +105,8 @@ addWebAuthorization(const char *url,
 	} else if (urlProx)
 		url = getDataURL(url);
 
-	host = getHostURL(url);
+	if (!getProtHostURL(url, NULL, host))
+		return false;
 	port = getPortURL(url);
 	if (!proxy) {
 		getDirURL(url, &dir, &dirend);
@@ -114,27 +120,29 @@ addWebAuthorization(const char *url,
 		    stringEqualCI(a->host, host) &&
 		    (proxy || (dl == strlen(a->directory)
 			       && !memcmp(a->directory, dir, dl)))) {
-			nzFree(a->user_password);
+			char *s = cloneString(credentials);
+			char *t = a->user_password;
+			a->user_password = s;
+			nzFree(t);
+			updated = true;
 			break;
 		}
 	}
 
-	if (a == (struct httpAuth *)&authlist) {
-		updated = false;
+	if (!updated) {
 		a = allocZeroMem(sizeof(struct httpAuth));
+		a->proxy = proxy;
+		a->port = port;
+		if (!a->host)
+			a->host = cloneString(host);
+		if (dir && !a->directory)
+			a->directory = pullString1(dir, dirend);
+		if (realm && !a->realm)
+			a->realm = cloneString(realm);
+		a->user_password = cloneString(credentials);
 		addToListFront(&authlist, a);
 	}
 
-	a->proxy = proxy;
-	a->port = port;
-	if (!a->host)
-		a->host = cloneString(host);
-	if (dir && !a->directory)
-		a->directory = pullString1(dir, dirend);
-	if (realm && !a->realm)
-		a->realm = cloneString(realm);
-
-	a->user_password = cloneString(credentials);
 	debugPrint(3, "%s authorization for %s%s",
 		   updated ? "updated" : "new", a->host, a->directory);
 	return true;
