@@ -136,6 +136,21 @@ struct htmlTag *findOpenTag(struct htmlTag *t, int action)
 	return 0;
 }				/* findOpenTag */
 
+static struct htmlTag *findOpenSection(struct htmlTag *t)
+{
+	int count = 0;
+	while ((t = t->parent)) {
+		if (t->action == TAGACT_TBODY || t->action == TAGACT_THEAD ||
+		    t->action == TAGACT_TFOOT)
+			return t;
+		if (++count == 10000) {	// tree shouldn't be this deep
+			debugPrint(1, "infinite loop in findOpenTag()");
+			break;
+		}
+	}
+	return 0;
+}				/* findOpenSection */
+
 struct htmlTag *findOpenList(struct htmlTag *t)
 {
 	while ((t = t->parent))
@@ -199,6 +214,83 @@ static void nestedAnchors(int start)
 	}
 }				/* nestedAnchors */
 
+/*********************************************************************
+Tables are suppose to have bodies, I guess.
+So <table><tr> becomes <table><tbody><tr>
+Find each table and look at its children.
+Note the tags between sections, where section is tHead, tBody, or tFoot.
+If that span includes <tr>, then put those tags under a new tBody.
+*********************************************************************/
+
+static void insert_tbody1(struct htmlTag *s1, struct htmlTag *s2,
+			  struct htmlTag *tbl);
+static bool tagBelow(struct htmlTag *t, int action);
+
+static void insert_tbody(int start)
+{
+	int i, end = cw->numTags;
+	struct htmlTag *tbl, *s1, *s2;
+
+	for (i = start; i < end; ++i) {
+		tbl = tagList[i];
+		if (tbl->action != TAGACT_TABLE)
+			continue;
+		s1 = 0;
+		do {
+			s2 = (s1 ? s1->sibling : tbl->firstchild);
+			while (s2 && s2->action != TAGACT_TBODY
+			       && s2->action != TAGACT_THEAD
+			       && s2->action != TAGACT_TFOOT)
+				s2 = s2->sibling;
+			insert_tbody1(s1, s2, tbl);
+			s1 = s2;
+		} while (s1);
+	}
+}
+
+static void insert_tbody1(struct htmlTag *s1, struct htmlTag *s2,
+			  struct htmlTag *tbl)
+{
+	struct htmlTag *s1a = (s1 ? s1->sibling : tbl->firstchild);
+	struct htmlTag *u, *uprev, *ns;	// new section
+
+	if (s1a == s2)		// nothing between
+		return;
+
+// Look for the direct html <table><tr><th>.
+// If th is anywhere else down the path, we won't find it.
+	if (!s1 && s1a->action == TAGACT_TR &&
+	    (u = s1a->firstchild) && stringEqual(u->info->name, "th")) {
+		ns = newTag("thead");
+		tbl->firstchild = ns;
+		ns->parent = tbl;
+		ns->firstchild = s1a;
+		s1a->parent = ns;
+		ns->sibling = s1a->sibling;
+		s1a->sibling = 0;
+		s1 = ns;
+		s1a = s1->sibling;
+	}
+
+	for (u = s1a; u != s2; u = u->sibling)
+		if (tagBelow(u, TAGACT_TR))
+			break;
+	if (u == s2)		// no rows below
+		return;
+
+	ns = newTag("tbody");
+	for (u = s1a; u != s2; u = u->sibling)
+		uprev = u, u->parent = ns;
+	if (s1)
+		s1->sibling = ns;
+	else
+		tbl->firstchild = ns;
+	if (s2)
+		uprev->sibling = 0, ns->sibling = s2;
+	ns->firstchild = s1a;
+	ns->parent = tbl;
+}
+
 /*********************************************************************
 Bad html will derail tidy, so that <a><div>stuff</div></a>
 will push div outside the anchor, to render as  {} stuff
@@ -228,7 +320,7 @@ innerHTML is wrong, and doesn't match the tree of nodes
 or the original source.
 innerHTML comes to us from tidy, after it has fixed (sometimes broken) things.
 Add <script> to the above, browse, jdb, and look at document.body.innerHTML.
-It does not match the source, in fact it represent the tree *before* we fixed it.
+It does not match the source, in fact it represents the tree *before* we fixed it.
 There really isn't anything I can do about that.
 In so many ways, the better approach is to fix tidy, but sometimes that is out of our hands.
 *********************************************************************/
@@ -355,9 +447,9 @@ void htmlInputHelper(struct htmlTag *t)
 	int len;
 	char *myname = (t->name ? t->name : t->id);
 	const char *s = attribVal(t, "type");
-	if (stringEqual(t->info->name, "button")) {
-		n = INP_BUTTON;
-	}
+	bool isbutton = stringEqual(t->info->name, "button");
+
+	t->itype = (isbutton ? INP_BUTTON : INP_TEXT);
 	if (s) {
 		n = stringInListCI(inp_types, s);
 		if (n < 0) {
@@ -365,13 +457,15 @@ void htmlInputHelper(struct htmlTag *t)
 			if (n < 0)
 				debugPrint(3, "unrecognized input type %s", s);
 			else
-				t->itype_minor = n;
+				t->itype = INP_TEXT, t->itype_minor = n;
 			if (n == INP_PW)
 				t->masked = true;
-			n = INP_TEXT;
-		}
+		} else
+			t->itype = n;
 	}
-	t->itype = n;
+// button no type means submit
+	if (!s && isbutton)
+		t->itype = INP_SUBMIT;
 
 	s = attribVal(t, "maxlength");
 	len = 0;
@@ -466,9 +560,9 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 			break;
 
 		if (currentTitle) {
-			if (!cw->ft) {
-				cw->ft = cloneString(t->textval);
-				spaceCrunch(cw->ft, true, false);
+			if (!cw->htmltitle) {
+				cw->htmltitle = cloneString(t->textval);
+				spaceCrunch(cw->htmltitle, true, false);
 			}
 			t->deleted = true;
 			break;
@@ -482,7 +576,6 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 		}
 
 		if (currentStyle) {
-			currentStyle->textval = cloneString(t->textval);
 			t->deleted = true;
 			break;
 		}
@@ -637,7 +730,6 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 			break;
 		}
 		currentStyle = t;
-		t->textval = emptyString;
 		break;
 
 	case TAGACT_SELECT:
@@ -686,11 +778,21 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 		}
 		break;
 
-	case TAGACT_TR:
+	case TAGACT_TBODY:
+	case TAGACT_THEAD:
+	case TAGACT_TFOOT:
 		if (opentag)
 			t->controller = findOpenTag(t, TAGACT_TABLE);
 		break;
 
+	case TAGACT_TR:
+		if (opentag) {
+			t->controller = findOpenSection(t);
+			if (!t->controller)
+				t->controller = findOpenTag(t, TAGACT_TABLE);
+		}
+		break;
+
 	case TAGACT_TD:
 		if (opentag)
 			t->controller = findOpenTag(t, TAGACT_TR);
@@ -734,9 +836,10 @@ static void prerenderNode(struct htmlTag *t, bool opentag)
 
 void prerender(int start)
 {
-/* some cleanup routines to rearrange the tree, working around some tidy5 bugs. */
+/* some cleanup routines to rearrange the tree */
 	nestedAnchors(start);
 	emptyAnchors(start);
+	insert_tbody(start);
 	tableForm(start);
 
 	currentForm = currentSel = currentOpt = NULL;
@@ -761,80 +864,8 @@ jsobjtype instantiate_url(jsobjtype parent, const char *name, const char *url)
 	return uo;
 }				/* instantiate_url */
 
-static void handlerSet(jsobjtype ev, const char *name, const char *code)
-{
-	char *newcode = allocMem(strlen(code) + 60);
-	strcpy(newcode, "(function(){");
-
-/*********************************************************************
-I read somewhere, long ago, that handlers are suppose to run in the context of
-the document object and sometimes the form that houses them.
-So I should put with(document) and sometimes with(form) around the code.
-For onclick, the code just runs, but for onsubmit the code is suppose to
-return true or false.
-Mozilla had no trouble compiling and running  return true;  at top level.
-Duktape won't do that. Return has to be in a function.
-I even tried DUK_COMPILE_FUNCTION, no dice.
-So I tried wrapping the code in
-(function() { code })();
-Then it doesn't matter if the code is just expression, or return expression.
-But should the with clauses be inside or outside the function?
-Also, the this binding was lost,
-which is absolutely necessary.
-Sami taught me how to get the binding back, via the .bind method,
-which is pretty dog gone cool.
-But what about the with clauses?
-Well so far I haven't found an example that needs with(document) around it,
-so I'm not going to worry about that right now.
-
-	strcpy(newcode, "with(document) { ");
-	enum ej_proptype hasform = typeof_property(ev, "form");
-	if (hasform)
-		strcat(newcode, "with(this.form) { ");
-*********************************************************************/
-
-	strcat(newcode, code);
-
-/*********************************************************************
-	if (hasform)
-		strcat(newcode, " }");
-	strcat(newcode, " }");
-*********************************************************************/
-
-	strcat(newcode, " }.bind(this))()");
-	set_property_function(ev, name, newcode);
-	nzFree(newcode);
-}				/* handlerSet */
-
-static void set_onhandler(const struct htmlTag *t, const char *name)
-{
-	const char *s;
-	if (t->jv) {
-		s = attribVal(t, name);
-		if (s)
-			handlerSet(t->jv, name, s);
-	}
-}				/* set_onhandler */
-
-static void set_onhandlers(const struct htmlTag *t)
-{
-/* I don't do anything with onkeypress, onfocus, etc,
- * these are just the most common handlers */
-	if (t->onclick)
-		set_onhandler(t, "onclick");
-	if (t->onchange)
-		set_onhandler(t, "onchange");
-	if (t->onsubmit)
-		set_onhandler(t, "onsubmit");
-	if (t->onreset)
-		set_onhandler(t, "onreset");
-	if (t->onload)
-		set_onhandler(t, "onload");
-	if (t->onunload)
-		set_onhandler(t, "onunload");
-}				/* set_onhandlers */
-
 static char fakePropLast[24];
+static jsobjtype fakePropParent;
 static const char *fakePropName(void)
 {
 	static int idx = 0;
@@ -848,6 +879,7 @@ static jsobjtype establish_js_option(jsobjtype obj, int idx)
 	jsobjtype oa;		/* option array */
 	jsobjtype oo;		/* option object */
 	jsobjtype so;		// style object
+	jsobjtype ato;		// attributes object
 	jsobjtype fo;		/* form object */
 
 	if ((oa = get_property_object(obj, "options")) == NULL)
@@ -862,7 +894,8 @@ static jsobjtype establish_js_option(jsobjtype obj, int idx)
 	if (fo)
 		set_property_object(oo, "form", fo);
 	instantiate_array(oo, "childNodes");
-	instantiate_array(oo, "attributes");
+	ato = instantiate(oo, "attributes", "NamedNodeMap");
+	set_property_object(ato, "owner", oo);
 	so = instantiate(oo, "style", "CSSStyleDeclaration");
 	set_property_object(so, "element", oo);
 
@@ -873,7 +906,7 @@ static void establish_inner(jsobjtype obj, const char *start, const char *end,
 			    bool isText)
 {
 	const char *s = emptyString;
-	const char *name = (isText ? "innerText" : "innerHTML");
+	const char *name = (isText ? "value" : "innerHTML");
 	if (start) {
 		s = start;
 		if (end)
@@ -882,13 +915,16 @@ static void establish_inner(jsobjtype obj, const char *start, const char *end,
 	set_property_string(obj, name, s);
 	if (start && end)
 		nzFree((char *)s);
+// If this is a textarea, we haven't yet set up the innerHTML
+// getter and seter
+	if (isText)
+		set_property_string(obj, "innerHTML", emptyString);
 }				/* establish_inner */
 
 static void domLink(struct htmlTag *t, const char *classname,	/* instantiate this class */
 		    const char *href, const char *list,	/* next member of this array */
-		    jsobjtype owner, int radiosel)
+		    jsobjtype owner, bool isradio)
 {
-	jsobjtype master;
 	jsobjtype alist = 0;
 	jsobjtype io = 0;	/* input object */
 	int length;
@@ -898,11 +934,14 @@ static void domLink(struct htmlTag *t, const char *classname,	/* instantiate thi
 	const char *idname = t->id;
 	const char *membername = 0;	/* usually symname */
 	const char *href_url = t->href;
+	const char *tcn = t->class;
 	const char *stylestring = attribVal(t, "style");
 	jsobjtype so = 0;	/* obj.style */
+	jsobjtype ato = 0;	/* obj.attributes */
+	char upname[MAXTAGNAME];
 
 	debugPrint(5, "domLink %s.%d name %s",
-		   classname, radiosel, (symname ? symname : emptyString));
+		   classname, isradio, (symname ? symname : emptyString));
 
 	if (symname && typeof_property(owner, symname)) {
 /*********************************************************************
@@ -916,11 +955,10 @@ The first time we create the array,
 and thereafter we just link under that array.
 Or - and this really does happen -
 an input tag could have the name action, colliding with form.action.
-I have no idea what to do here.
-radiosel is 1 for radio buttons and 2 for select.
+don't overwrite form.action, or anything else that pre-exists.
 *********************************************************************/
 
-		if (radiosel == 1) {
+		if (isradio) {
 /* name present and radio buttons, name should be the array of buttons */
 			io = get_property_object(owner, symname);
 			if (io == NULL)
@@ -940,55 +978,45 @@ Ok, the above condition does not hold.
 We'll be creating a new object under owner, but through what name?
 The name= tag, unless it's a duplicate,
 or id= if there is no name=, or a fake name just to protect it from gc.
+That's how it was for a long time, but I think we only do this on form.
 *********************************************************************/
-
-		if (!symname && idname) {
-			membername = idname;
-		} else if (symname && !dupname) {
-			membername = symname;
-		}
+		if (t->action == TAGACT_INPUT && list) {
+			if (!symname && idname)
+				membername = idname;
+			else if (symname && !dupname)
+				membername = symname;
 /* id= or name= must not displace submit, reset, or action in form.
  * Example www.startpage.com, where id=submit.
  * nor should it collide with another attribute, such as document.cookie and
  * <div ID=cookie> in www.orange.com.
  * This call checks for the name in the object and its prototype. */
-		if (has_property(owner, membername)) {
-			debugPrint(3, "membername overload %s.%s", classname,
-				   membername);
-			membername = NULL;
+			if (membername && has_property(owner, membername)) {
+				debugPrint(3, "membername overload %s.%s",
+					   classname, membername);
+				membername = NULL;
+			}
 		}
-		if (!membername)
+		if (!membername) {
 			membername = fakePropName();
+			fakePropParent = owner;
+		}
 
-		if (radiosel) {
-/* The first radio button, or input type=select */
-/* Either way the form element is suppose to be an array. */
+		if (isradio) {	// the first radio button
 			io = instantiate_array(owner, membername);
 			if (io == NULL)
 				return;
-			if (radiosel == 1) {
-				set_property_string(io, "type", "radio");
-				set_property_string(io, "nodeName", "radio");
-			} else {
-/* I've read some docs that say select is itself an array,
- * and then references itself as an array of options.
- * Self referencing? Really? Well it seems to work. */
-				set_property_object(io, "options", io);
-				set_property_object(io, "childNodes", io);
-				set_property_number(io, "selectedIndex", -1);
-			}
+			set_property_string(io, "type", "radio");
 		} else {
 /* A standard input element, just create it. */
+			jsobjtype ca;	// child array
 			io = instantiate(owner, membername, classname);
 			if (io == NULL)
 				return;
-
 /* not an array; needs the childNodes array beneath it for the children */
-			instantiate_array(io, "childNodes");
-
-/* in the special case of form, also need an array of elements */
-			if (t->action == TAGACT_FORM)
-				instantiate_array(io, "elements");
+			ca = instantiate_array(io, "childNodes");
+// childNodes and options are the same for Select
+			if (stringEqual(classname, "Select"))
+				set_property_object(io, "options", ca);
 		}
 
 /* deal with the 'styles' here.
@@ -1009,21 +1037,18 @@ Don't do any of this if the tag is itself <style>. */
 
 /* Other attributes that are expected by pages, even if they
  * aren't populated at domLink-time */
-		set_property_string(io, "class", "");
-		set_property_string(io, "nodeValue", "");
-		instantiate_array(io, "attributes");
+		if (!tcn)
+			tcn = emptyString;
+		set_property_string(io, "class", tcn);
+		set_property_string(io, "last$class", tcn);
+		ato = instantiate(io, "attributes", "NamedNodeMap");
+		set_property_object(ato, "owner", io);
 		set_property_object(io, "ownerDocument", cf->docobj);
 
-		if (membername == symname) {
-/* link to document.all */
-			master = get_property_object(cf->docobj, "all");
-			if (master == NULL)
-				return;
-			set_property_object(master, symname, io);
-
-			if (stringEqual(symname, "action"))
-				set_property_bool(io, "actioncrash", true);
-		}
+// only anchors with href go into links[]
+		if (list && stringEqual(list, "links") &&
+		    !attribPresent(t, "href"))
+			list = 0;
 
 		if (list)
 			alist = get_property_object(owner, list);
@@ -1035,13 +1060,15 @@ Don't do any of this if the tag is itself <style>. */
 			if (symname && !dupname
 			    && !has_property(alist, symname))
 				set_property_object(alist, symname, io);
+#if 0
 			if (idname && symname != idname
 			    && !has_property(alist, idname))
 				set_property_object(alist, idname, io);
+#endif
 		}		/* list indicated */
 	}
 
-	if (radiosel == 1) {
+	if (isradio) {
 /* drop down to the element within the radio array, and return that element */
 /* w becomes the object associated with this radio button */
 /* io is, by assumption, an array */
@@ -1058,12 +1085,8 @@ Don't do any of this if the tag is itself <style>. */
 	if (symname)
 		set_property_string(io, "name", symname);
 
-	if (idname) {
-/* io.id becomes idname, and idMaster.idname becomes io */
+	if (idname)
 		set_property_string(io, "id", idname);
-		master = get_property_object(cf->docobj, "idMaster");
-		set_property_object(master, idname, io);
-	}
 
 	if (href && href_url)
 // This use to be instantiate_url, but with the new side effects
@@ -1077,19 +1100,11 @@ Don't do any of this if the tag is itself <style>. */
 
 	connectTagObject(t, io);
 
-	set_property_string(io, "nodeName", t->info->name);
-/* documentElement is now set in the "Body" case because the 
-"Html" does not appear ever to be encountered */
-
-	if (t->action == TAGACT_BODY) {
-		set_property_object(cf->docobj, "body", io);
-		set_property_object(cf->docobj, "documentElement", io);
-	}
-
-	if (t->action == TAGACT_HEAD) {
-		set_property_object(cf->docobj, "head", io);
-	}
-
+	strcpy(upname, t->info->name);
+	caseShift(upname, 'u');
+	set_property_string(io, "nodeName", upname);
+	set_property_string(io, "tagName", upname);
+	set_property_number(io, "nodeType", 1);
 }				/* domLink */
 
 static const char defvl[] = "defaultValue";
@@ -1100,21 +1115,19 @@ static void formControlJS(struct htmlTag *t)
 {
 	const char *typedesc;
 	int itype = t->itype;
-	int isradio = itype == INP_RADIO;
-	int isselect = (itype == INP_SELECT) * 2;
+	bool isradio = (itype == INP_RADIO);
+	bool isselect = (itype == INP_SELECT);
+	const char *whichclass = (isselect ? "Select" : "Element");
 	const struct htmlTag *form = t->controller;
 
 	if (form && form->jv)
-		domLink(t, "Element", 0, "elements", form->jv,
-			isradio | isselect);
+		domLink(t, whichclass, 0, "elements", form->jv, isradio);
 	else
-		domLink(t, "Element", 0, 0, cf->docobj, isradio | isselect);
+		domLink(t, whichclass, 0, 0, cf->docobj, isradio);
 	if (!t->jv)
 		return;
 
-	set_onhandlers(t);
-
-	if (itype <= INP_RADIO) {
+	if (itype <= INP_RADIO && !isselect) {
 		set_property_string(t->jv, "value", t->value);
 		if (itype != INP_FILE) {
 /* No default value on file, for security reasons */
@@ -1138,6 +1151,7 @@ static void optionJS(struct htmlTag *t)
 {
 	struct htmlTag *sel = t->controller;
 	const char *tx = t->textval;
+	const char *cl = t->class;
 
 	if (!sel)
 		return;
@@ -1156,64 +1170,88 @@ static void optionJS(struct htmlTag *t)
 	connectTagObject(t, establish_js_option(sel->jv, t->lic));
 	set_property_string(t->jv, "text", t->textval);
 	set_property_string(t->jv, "value", t->value);
-	set_property_string(t->jv, "nodeName", "option");
+	set_property_string(t->jv, "nodeName", "OPTION");
+	set_property_number(t->jv, "nodeType", 1);
 	set_property_bool(t->jv, "selected", t->checked);
 	set_property_bool(t->jv, defsel, t->checked);
+	if (!cl)
+		cl = emptyString;
+	set_property_string(t->jv, "class", cl);
+	set_property_string(t->jv, "last$class", cl);
 
-	if (t->checked && !sel->multiple) {
+	if (t->checked && !sel->multiple)
 		set_property_number(sel->jv, "selectedIndex", t->lic);
-		set_property_string(sel->jv, "value", t->value);
-	}
 }				/* optionJS */
 
 static void link_css(struct htmlTag *t)
 {
-	const char *a = attribVal(t, "type");
-	if (a)
-		set_property_string(t->jv, "type", a);
+	struct i_get g;
+	char *b;
+	int blen;
+	const char *a;
+	const char *a1 = attribVal(t, "type");
+	const char *a2 = attribVal(t, "rel");
+	if (a1)
+		set_property_string(t->jv, "type", a1);
+	if (a2)
+		set_property_string(t->jv, "rel", a2);
 	if (!t->href)
 		return;
-	if (!a || !stringEqualCI(a, "text/css"))
+	if ((!a1 || !stringEqualCI(a1, "text/css")) &&
+	    (!a2 || !stringEqualCI(a2, "stylesheet")))
 		return;
 // Fetch the css file so we can apply its attributes.
 	a = NULL;
 	if (browseLocal && !isURL(t->href)) {
 		debugPrint(3, "css source %s", t->href);
-		if (!fileIntoMemory(t->href, &serverData, &serverDataLen)) {
+		if (!fileIntoMemory(t->href, &b, &blen)) {
 			if (debugLevel >= 1)
-				i_printf(MSG_GetLocalCSS, errorMsg);
+				i_printf(MSG_GetLocalCSS);
 		} else {
-			a = force_utf8(serverData, serverDataLen);
+			a = force_utf8(b, blen);
 			if (!a)
-				a = serverData;
+				a = b;
 			else
-				nzFree(serverData);
-			serverData = NULL;
-			serverDataLen = 0;
+				nzFree(b);
 		}
 	} else {
 		debugPrint(3, "css source %s", t->href);
-		if (httpConnect(t->href, false, false, true, 0, 0, 0)) {
-			if (ht_code == 200) {
-				a = force_utf8(serverData, serverDataLen);
+		memset(&g, 0, sizeof(g));
+		g.thisfile = cf->fileName;
+		g.uriEncoded = true;
+		g.url = t->href;
+		if (httpConnect(&g)) {
+			if (g.code == 200) {
+				a = force_utf8(g.buffer, g.length);
 				if (!a)
-					a = serverData;
+					a = g.buffer;
 				else
-					nzFree(serverData);
+					nzFree(g.buffer);
+// acid3 test[0] says we don't process this file if it's content type is
+// text/html. Should I test for anything outside of text/css?
+// For now I insist it be missing or text/css or text/plain.
+// A similar test is performed in css.c after httpConnect.
+				if (g.content[0]
+				    && !stringEqual(g.content, "text/css")
+				    && !stringEqual(g.content, "text/plain")) {
+					debugPrint(3,
+						   "css suppressed because content type is %s",
+						   g.content);
+					cnzFree(a);
+					a = NULL;
+				}
 			} else {
-				nzFree(serverData);
+				nzFree(g.buffer);
 				if (debugLevel >= 3)
-					i_printf(MSG_GetCSS, t->href, ht_code);
+					i_printf(MSG_GetCSS, g.url, g.code);
 			}
-			serverData = NULL;
-			serverDataLen = 0;
 		} else {
 			if (debugLevel >= 3)
-				i_printf(MSG_GetCSS2, errorMsg);
+				i_printf(MSG_GetCSS2);
 		}
 	}
 	if (a)
-		set_property_string(t->jv, "data", a);
+		set_property_string(t->jv, "css$data", a);
 	cnzFree(a);
 }				/* link_css */
 
@@ -1227,6 +1265,10 @@ static void jsNode(struct htmlTag *t, bool opentag)
 	const char *a;
 	bool linked_in;
 
+// run reindex at table close
+	if (action == TAGACT_TABLE && !opentag && t->jv)
+		run_function_onearg(cf->winobj, "rowReindex", t->jv);
+
 /* all the js variables are on the open tag */
 	if (!opentag)
 		return;
@@ -1252,19 +1294,25 @@ Needless to say that's not good!
 		connectTagObject(t,
 				 instantiate(cf->docobj, fakePropName(),
 					     "TextNode"));
+// nodeName and nodeType set in constructor
 		if (t->jv) {
 			const char *w = t->textval;
 			if (!w)
 				w = emptyString;
 			set_property_string(t->jv, "data", w);
+			w = (t->class ? t->class : emptyString);
+			set_property_string(t->jv, "class", w);
+			set_property_string(t->jv, "last$class", w);
 		}
 		break;
 
+	case TAGACT_HTML:
+		domLink(t, "HTML", 0, 0, cf->docobj, 0);
+		cf->htmltag = t;
+		break;
+
 	case TAGACT_META:
 		domLink(t, "Meta", 0, "metas", cf->docobj, 0);
-		a = attribVal(t, "content");
-		set_property_string(t->jv, "content", a);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_STYLE:
@@ -1273,9 +1321,6 @@ Needless to say that's not good!
 		if (!a)
 			a = emptyString;
 		set_property_string(t->jv, "type", a);
-		set_property_number(t->jv, "nodeType", 1);
-		if (t->textval && t->textval[0])
-			set_property_string(t->jv, "data", t->textval);
 		break;
 
 	case TAGACT_SCRIPT:
@@ -1283,9 +1328,6 @@ Needless to say that's not good!
 		a = attribVal(t, "type");
 		if (a)
 			set_property_string(t->jv, "type", a);
-		a = attribVal(t, "language");
-		if (a)
-			set_property_string(t->jv, "language", a);
 		a = attribVal(t, "src");
 		if (a) {
 			set_property_string(t->jv, "src", a);
@@ -1298,44 +1340,35 @@ Needless to say that's not good!
 		} else {
 			set_property_string(t->jv, "data", "");
 		}
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_FORM:
 		domLink(t, "Form", "action", "forms", cf->docobj, 0);
-		set_onhandlers(t);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_INPUT:
 		formControlJS(t);
 		if (t->itype == INP_TA)
 			establish_inner(t->jv, t->value, 0, true);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_OPTION:
 		optionJS(t);
 // The parent child relationship has already been established,
 // don't break, just return;
-		set_property_number(t->jv, "nodeType", 1);
 		return;
 
 	case TAGACT_A:
-		domLink(t, "Anchor", "href", "anchors", cf->docobj, 0);
-		set_onhandlers(t);
-		set_property_number(t->jv, "nodeType", 1);
+		domLink(t, "Anchor", "href", "links", cf->docobj, 0);
 		break;
 
 	case TAGACT_HEAD:
 		domLink(t, "Head", 0, "heads", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
+		cf->headtag = t;
 		break;
 
 	case TAGACT_BODY:
 		domLink(t, "Body", 0, "bodies", cf->docobj, 0);
-		set_onhandlers(t);
-		set_property_number(t->jv, "nodeType", 1);
 		cf->bodytag = t;
 		break;
 
@@ -1343,44 +1376,59 @@ Needless to say that's not good!
 	case TAGACT_UL:
 	case TAGACT_DL:
 		domLink(t, "Lister", 0, 0, cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_LI:
 		domLink(t, "Listitem", 0, 0, cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TABLE:
 		domLink(t, "Table", 0, "tables", cf->docobj, 0);
-/* create the array of rows under the table */
-		instantiate_array(t->jv, "rows");
-		set_property_number(t->jv, "nodeType", 1);
+		break;
+
+	case TAGACT_TBODY:
+		if ((above = t->controller) && above->jv)
+			domLink(t, "tBody", 0, "tBodies", above->jv, 0);
+		break;
+
+	case TAGACT_THEAD:
+		if ((above = t->controller) && above->jv) {
+			domLink(t, "tHead", 0, 0, above->jv, 0);
+			set_property_object(above->jv, "tHead", t->jv);
+		}
+		break;
+
+	case TAGACT_TFOOT:
+		if ((above = t->controller) && above->jv) {
+			domLink(t, "tFoot", 0, 0, above->jv, 0);
+			set_property_object(above->jv, "tFoot", t->jv);
+		}
 		break;
 
 	case TAGACT_TR:
-		if ((above = t->controller) && above->jv) {
-			domLink(t, "Trow", 0, "rows", above->jv, 0);
-			instantiate_array(t->jv, "cells");
-		}
-		set_property_number(t->jv, "nodeType", 1);
+		if ((above = t->controller) && above->jv)
+			domLink(t, "tRow", 0, "rows", above->jv, 0);
 		break;
 
 	case TAGACT_TD:
-		if ((above = t->controller) && above->jv) {
+		if ((above = t->controller) && above->jv)
 			domLink(t, "Cell", 0, "cells", above->jv, 0);
-		}
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_DIV:
 		domLink(t, "Div", 0, "divs", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
+		break;
+
+	case TAGACT_LABEL:
+		domLink(t, "Label", 0, "labels", cf->docobj, 0);
 		break;
 
 	case TAGACT_OBJECT:
 		domLink(t, "HtmlObj", 0, "htmlobjs", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
+		break;
+
+	case TAGACT_UNKNOWN:
+		domLink(t, "HTMLElement", 0, 0, cf->docobj, 0);
 		break;
 
 	case TAGACT_SPAN:
@@ -1388,12 +1436,10 @@ Needless to say that's not good!
 	case TAGACT_SUP:
 	case TAGACT_OVB:
 		domLink(t, "Span", 0, "spans", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_AREA:
-		domLink(t, "Area", "href", "areas", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
+		domLink(t, "Area", "href", "links", cf->docobj, 0);
 		break;
 
 	case TAGACT_FRAME:
@@ -1403,39 +1449,32 @@ Needless to say that's not good!
 			t->href = 0;
 		}
 		domLink(t, "Frame", "src", "frames", cf->winobj, 0);
-		set_onhandlers(t);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_IMAGE:
 		domLink(t, "Image", "src", "images", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_P:
 		domLink(t, "P", 0, "paragraphs", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_HEADER:
 		domLink(t, "Header", 0, "headers", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_FOOTER:
 		domLink(t, "Footer", 0, "footers", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
 		break;
 
 	case TAGACT_TITLE:
-		if (cw->ft)
-			set_property_string(cf->docobj, "title", cw->ft);
-// Title is not a node, more like an attribute of the page.
+		if (cw->htmltitle)
+			set_property_string(cf->docobj, "title", cw->htmltitle);
+		domLink(t, "Title", 0, 0, cf->docobj, 0);
 		break;
 
 	case TAGACT_LINK:
-		domLink(t, "Link", "href", "links", cf->docobj, 0);
-		set_property_number(t->jv, "nodeType", 1);
+		domLink(t, "Link", "href", 0, cf->docobj, 0);
 		link_css(t);
 		break;
 
@@ -1452,34 +1491,42 @@ Needless to say that's not good!
 		return;		/* nothing else to do */
 
 /* js tree mirrors the dom tree. */
-/* but head and body link to document */
 	linked_in = false;
-	if (action == TAGACT_HEAD || action == TAGACT_BODY) {
+
+	if (t->parent && t->parent->jv) {
+		run_function_onearg(t->parent->jv, "eb$apch1", t->jv);
+		linked_in = true;
+// special code for frame.contentDocument.
+		if (t->parent->action == TAGACT_FRAME) {
+			set_property_object(t->parent->jv,
+					    "contentDocument", t->jv);
+			set_property_object(t->parent->jv,
+					    "contentWindow", t->jv);
+		}
+	}
+
+	if (action == TAGACT_HTML) {
 		run_function_onearg(cf->docobj, "eb$apch1", t->jv);
 		linked_in = true;
-	} else {
-		if (t->parent && t->parent->jv) {
-			run_function_onearg(t->parent->jv, "eb$apch1", t->jv);
-			linked_in = true;
-// special code for frame.contentDocument.
-			if (t->parent->action == TAGACT_FRAME) {
-				set_property_object(t->parent->jv,
-						    "contentDocument", t->jv);
-				set_property_object(t->parent->jv,
-						    "contentWindow", t->jv);
-			}
-		}
+	}
 
-		if (!t->parent && innerParent) {
-			run_function_onearg(innerParent, "eb$apch1", t->jv);
-			linked_in = true;
-		}
+	if (!t->parent && innerParent) {
+// this is the top of innerHTML or some such.
+// It is never html head or body, as those are skipped.
+		run_function_onearg(innerParent, "eb$apch1", t->jv);
+		linked_in = true;
 	}
 
 	if (linked_in && fakePropLast[0]) {
 // Node linked to document/gc to protect if from garbage collection,
 // but now it is linked to its parent.
-		delete_property(cf->docobj, fakePropLast);
+		delete_property(fakePropParent, fakePropLast);
+	}
+
+	if (!linked_in) {
+		debugPrint(3, "tag %s not linked in", ti->name);
+		if (action == TAGACT_TEXT)
+			debugPrint(1, "text %s\n", t->textval);
 	}
 
 /* set innerHTML from the source html, if this tag supports it */
@@ -1499,21 +1546,29 @@ static void pushAttributes(const struct htmlTag *t)
 		return;
 	for (i = 0; a[i]; ++i) {
 // There are some exceptions, some attributes that we handle individually.
-		const char *const exclist[] = {
-			"onclick", "onchange", "onsubmit", "onreset", "onload",
-			"onunload",
-			"name", "id",
+		static const char *const exclist[] = {
+			"name", "id", "class",
 			"checked", "value", "type", "style",
 			"href", "src", "action",
 			0
 		};
+		static const char *const dotrue[] = {
+			"multiple", "readonly", "disabled", 0
+		};
+		static const char *const handlers[] = {
+			"onload", "onunload", "onclick", "onchange",
+			"onsubmit", "onreset",
+			0
+		};
 		const char *u;
 		if (stringInListCI(exclist, a[i]) >= 0)
 			continue;
 // I surely haven't thought of everything, so check generally.
 // Maybe they wrote <a firstChild=foo>
-// See if the name is specifically in the prototype.
-		if (has_property(t->jv, a[i]) && !typeof_property(t->jv, a[i])) {
+// See if the name is in the prototype, and not a handler,
+// as handlers have setters.
+		if (has_property(t->jv, a[i]) && !typeof_property(t->jv, a[i])
+		    && stringInList(handlers, a[i]) < 0) {
 			debugPrint(3, "html attribute overload %s.%s",
 				   t->info->name, a[i]);
 			continue;
@@ -1522,10 +1577,12 @@ static void pushAttributes(const struct htmlTag *t)
 		u = v[i];
 		if (!u)
 			u = emptyString;
-// There may be some, like multiple or readonly, that should be set to true,
-// not the empty string. I haven't dealt with those yet.
-// Should I call setAttribute, with its side effects, or is just setting
-// the property good enough??
+// There are some, like multiple or readonly, that should be set to true,
+// not the empty string.
+		if (!*u && stringInList(dotrue, a[i]) >= 0) {
+			set_property_bool(t->jv, a[i], true);
+			continue;
+		}
 		set_property_string(t->jv, a[i], u);
 	}
 }				/* pushAttributes */
@@ -1615,9 +1672,10 @@ void tag_gc(void)
 
 /* first three have to be in this order */
 const struct tagInfo availableTags[] = {
-	{"html", "html", TAGACT_ZERO},
+	{"html", "html", TAGACT_HTML},
 	{"base", "base reference for relative URLs", TAGACT_BASE, 0, 4},
-	{"object", "an html object", TAGACT_OBJECT, 5, 1},
+	{"unknown0", "an html entity", TAGACT_UNKNOWN, 5, 1},
+	{"object", "an html object", TAGACT_OBJECT, 5, 3},
 	{"a", "an anchor", TAGACT_A, 0, 1},
 	{"input", "an input item", TAGACT_INPUT, 0, 4},
 	{"element", "an input element", TAGACT_INPUT, 0, 4},
@@ -1650,7 +1708,8 @@ const struct tagInfo availableTags[] = {
 	{"div", "a divided section", TAGACT_DIV, 5, 1},
 	{"map", "a map of images", TAGACT_NOP, 5, 0},
 	{"blockquote", "a quoted paragraph", TAGACT_NOP, 10, 1},
-	{"document", "a document", TAGACT_NOP, 5, 1},
+	{"document", "a document", TAGACT_DOC, 5, 1},
+	{"comment", "a comment", TAGACT_COMMENT, 0, 2},
 	{"h1", "a level 1 header", TAGACT_NOP, 10, 1},
 	{"h2", "a level 2 header", TAGACT_NOP, 10, 1},
 	{"h3", "a level 3 header", TAGACT_NOP, 10, 1},
@@ -1667,13 +1726,15 @@ const struct tagInfo availableTags[] = {
 	{"dl", "a definition list", TAGACT_DL, 10, 1},
 	{"hr", "a horizontal line", TAGACT_HR, 5, 4},
 	{"form", "a form", TAGACT_FORM, 10, 1},
-	{"button", "a button", TAGACT_INPUT, 0, 0},
+	{"button", "a button", TAGACT_INPUT, 0, 1},
 	{"frame", "a frame", TAGACT_FRAME, 2, 0},
 	{"iframe", "a frame", TAGACT_FRAME, 2, 1},
 	{"map", "an image map", TAGACT_MAP, 2, 4},
 	{"area", "an image map area", TAGACT_AREA, 0, 4},
 	{"table", "a table", TAGACT_TABLE, 10, 1},
 	{"tbody", "a table body", TAGACT_TBODY, 0, 1},
+	{"thead", "a table body", TAGACT_THEAD, 0, 1},
+	{"tfoot", "a table body", TAGACT_TFOOT, 0, 1},
 	{"tr", "a table row", TAGACT_TR, 5, 1},
 	{"td", "a table entry", TAGACT_TD, 0, 5},
 	{"th", "a table heading", TAGACT_TD, 0, 5},
@@ -1691,13 +1752,15 @@ const struct tagInfo availableTags[] = {
 	{"embed", "embedded html", TAGACT_MUSIC, 0, 4},
 	{"noembed", "no embed section", TAGACT_NOP, 0, 2},
 	{"em", "emphasized text", TAGACT_JS, 0, 0},
-	{"label", "a label", TAGACT_JS, 0, 0},
+	{"label", "a label", TAGACT_LABEL, 0, 0},
 	{"strike", "emphasized text", TAGACT_JS, 0, 0},
 	{"s", "emphasized text", TAGACT_JS, 0, 0},
 	{"strong", "emphasized text", TAGACT_JS, 0, 0},
 	{"b", "bold text", TAGACT_JS, 0, 0},
 	{"i", "italicized text", TAGACT_JS, 0, 0},
 	{"u", "underlined text", TAGACT_JS, 0, 0},
+	{"var", "variable text", TAGACT_JS, 0, 0},
+	{"kbd", "keyboard text", TAGACT_JS, 0, 0},
 	{"dfn", "definition text", TAGACT_JS, 0, 0},
 	{"q", "quoted text", TAGACT_JS, 0, 0},
 	{"abbr", "an abbreviation", TAGACT_JS, 0, 0},
@@ -1712,10 +1775,11 @@ static void freeTag(struct htmlTag *t)
 	nzFree(t->textval);
 	nzFree(t->name);
 	nzFree(t->id);
+	nzFree(t->class);
+	nzFree(t->nodeName);
 	nzFree(t->value);
 	cnzFree(t->rvalue);
 	nzFree(t->href);
-	nzFree(t->class);
 	nzFree(t->js_file);
 	nzFree(t->innerHTML);
 
@@ -1778,7 +1842,6 @@ struct htmlTag *newTag(const char *name)
 {
 	struct htmlTag *t;
 	const struct tagInfo *ti;
-	int action;
 
 	for (ti = availableTags; ti->name[0]; ++ti)
 		if (stringEqualCI(ti->name, name))
@@ -1790,14 +1853,12 @@ struct htmlTag *newTag(const char *name)
 		ti = availableTags + 2;
 	}
 
-	if ((action = ti->action) == TAGACT_ZERO)
-		return 0;
-
 	t = (struct htmlTag *)allocZeroMem(sizeof(struct htmlTag));
+	t->action = ti->action;
 	t->f0 = cf;		/* set current frame */
-	t->action = action;
 	t->info = ti;
 	t->seqno = cw->numTags;
+	t->nodeName = cloneString(name);
 	pushTag(t);
 	return t;
 }				/* newTag */
@@ -1880,7 +1941,7 @@ static void intoTree(struct htmlTag *parent)
 				treeDisable = false;
 				continue;
 			}
-			if (action == TAGACT_BODY) {
+			if (action == TAGACT_HTML || action == TAGACT_BODY) {
 				debugPrint(4, "node pass %s", t->info->name);
 				t->dead = true;
 				++cw->deadTags;
@@ -2134,15 +2195,11 @@ static void processStyles(jsobjtype so, const char *stylestring)
 			trimWhite(sv);
 // the property name has to be nonempty
 			if (*s) {
-// foo-bar has to become fooBar
-				char *t, *w;
-				for (t = w = s; *t; ++t)
-					if (*t == '-' && isalpha(t[1]))
-						t[1] = toupper(t[1]);
-					else
-						*w++ = *t;
-				*w = 0;
+				cssAttributeCrunch(s);
 				set_property_string(so, s, sv);
+// Should we set a specification level here, perhaps high,
+// so the css sheets don't overwrite it?
+// sv + "$$scy" = 99999;
 			}
 		}
 	}
