@@ -30,7 +30,6 @@
 #include <sys/stat.h>
 #include <ctype.h>
 #include <string.h>
-#include <signal.h>
 #include <memory.h>
 #include <stdarg.h>
 #include <stdlib.h>
@@ -78,14 +77,6 @@ typedef uchar bool;
 #define true 1
 #endif
 
-/* Some source files are shared between edbrowse, a C program,
- * and edbrowse-js, currently a C++ function program, thus the prototypes,
- * and some other structures, must be C protected. */
-#ifdef __cplusplus
-// Because of this line, you can't meaningfully run this file through indent.
-extern "C" {
-#endif
-
 // Opaque indicator of an object that can be shared
 // between edbrowse and the js engine.
 typedef void *jsobjtype;
@@ -141,6 +132,8 @@ typedef uchar *pst;		/* perl string */
 #define ABSPATH 1024 // max length of an absolute pathname
 #define MAXRE 512 // max length of a regular expression
 #define MAXTTYLINE 256 // max length of an entered line
+#define MAXHOSTLEN 400
+#define MAXPROTLEN 12
 #define MAXUSERPASS 40 // user name or password
 #define MAXACCOUNT 100 // number of email accounts
 #define MAXAGENT 50 // number of user agents
@@ -187,16 +180,54 @@ extern bool newloc_r; /* location replaces this page */
 extern struct ebFrame *newloc_f; /* frame calling for new web page */
 extern const char *ebrc_string; /* default ebrc file */
 
-struct eb_curl_callback_data {
-// where to put the captured data.
-	char **buffer;
-	int *length;
+// Get data from the internet. Zero the structure, set the
+// members you need, then call httpConnect.
+struct i_get {
+// the data returned from the internet fetch
+	char *buffer;
+	int length;
+// in case you want the headers
+	char **headers_p;
+	const char *url;
+	char *urlcopy;
+	int urlcopy_l;
+	char *cfn; // changed filename
+	const char *thisfile;
+	char *referrer;
+	CURL *h;
 // State of download to disk, see http.c for state values.
 	int down_state;
 	int down_fd;	/* downloading file descriptor */
+	int down_msg;
 	const char *down_file;	/* downloading filename */
 	const char *down_file2;	/* without download directory */
 	int down_length;
+	bool down_ok;
+	bool uriEncoded;
+	bool foreground;
+	bool pg_ok; // watch for plugins
+	bool playonly; // only player plugins
+	bool csp; // content supresses plugins
+	bool is_http;
+	bool cacheable;
+	bool last_curlin;
+	bool move_capable;
+	char error[CURL_ERROR_SIZE + 1];
+	long code;		/* example, 404 */
+/* an assortment of variables that are gleaned from the incoming http headers */
+	char *headers;
+	int headers_len;
+/* http content type is used in many places, and isn't arbitrarily long
+ * or case sensitive, so keep our own sanitized copy. */
+	char content[60];
+	char *charset;	/* extra content info such as charset */
+int hcl;			/* http content length */
+	char *cdfn;		/* http content disposition file name */
+	time_t modtime;	/* http modification time */
+	char *etag;		/* the etag in the header */
+	char auth_realm[60];	/* WWW-Authenticate realm header */
+	char *newloc;
+	int newloc_d;
 };
 
 struct MACCOUNT {		/* pop3 account */
@@ -223,7 +254,7 @@ struct MIMETYPE {
 	char *urlmatch;
 	char *content;
 	char outtype;
-	bool stream, download;
+	bool down_url, from_file;
 };
 extern struct MIMETYPE mimetypes[];
 extern int maxMime;		/* how many mime types specified */
@@ -242,8 +273,9 @@ struct DBTABLE {
 extern CURL *global_http_handle;
 extern CURLSH *global_share_handle;
 extern int debugLevel;		/* 0 to 9 */
-extern bool debugClone, debugEvent, debugThrow;
+extern bool debugClone, debugEvent, debugThrow, debugCSS;
 extern bool demin; // deminimize javascript
+extern char *uvw; // undefined variable watch
 extern bool gotimers; // run javascript timers
 extern FILE *debugFile;
 extern char *debugFileName;
@@ -275,7 +307,6 @@ extern bool curlAuthNegotiate;  /* try curl negotiate (SPNEGO) auth */
 extern bool listNA;		/* list nonascii chars */
 extern bool inInput;		/* reading line from standard in */
 extern int fileSize;		/* when reading/writing files */
-extern long ht_code;		/* http code, like 404 file not found */
 extern char errorMsg[];		/* generated error message */
 extern int localAccount;	/* this is the smtp server for outgoing mail */
 extern char *mailDir;		/* move to this directory when fetching mail */
@@ -289,7 +320,7 @@ extern char *ebTempDir;		/* edbrowse temp, such as /tmp/.edbrowse */
 extern char *ebUserDir;		/* $ebTempDir/nnn user ID appended */
 extern char *dbarea, *dblogin, *dbpw;	/* to log into the database */
 extern bool fetchBlobColumns;
-extern bool caseInsensitive, searchStringsAll;
+extern bool caseInsensitive, searchStringsAll, searchWrap;
 extern bool allowRedirection;	/* from http code 301, or http refresh */
 extern bool sendReferrer;	/* in the http header */
 extern bool allowJS;		/* javascript on */
@@ -299,6 +330,7 @@ extern bool ftpActive;
 extern bool helpMessagesOn;	/* no need to type h */
 extern bool pluginsOn;		/* plugins are active */
 extern bool showHiddenFiles;	/* during directory scan */
+extern bool showHover; // messages that appear when you hover
 extern int context;		/* which session (buffer) are we in? */
 extern pst linePending;
 extern char *changeFileName;
@@ -306,7 +338,6 @@ extern char *addressFile;	/* your address book */
 extern char *serverData;
 extern int serverDataLen;
 extern char *breakLineResult;
-extern char *currentReferrer;
 extern char *home;		/* home directory */
 extern char *recycleBin;	/* holds deleted files */
 extern char *configFile, *sigFile, *sigFileEnd;
@@ -349,13 +380,15 @@ struct ebFrame {
 	char *fileName;		/* name of file or url */
 	char *firstURL;		/* before http redirection */
 	char *hbase; /* base for href references */
+	bool render1; // rendered via protocol or urlmatch
+	bool render2; // rendered via suffix
+	bool render1b;
 	bool baseset; // <base> tag has been seen
-	bool f_encoded; // filename is url encoded
-	bool fromcurl; // data comes from the internet
+	bool uriEncoded; // filename is url encoded
 	char *dw;		/* document.write string */
 	int dw_l;		/* length of the above */
 // document.writes go under the body.
-	struct htmlTag *bodytag;
+	struct htmlTag *htmltag, *headtag, *bodytag;
 /* The javascript context and window corresponding to this url or frame.
  * If this is null then javascript is not operational for this frame.
  * We could still be browsing however, without javascript. */
@@ -363,9 +396,20 @@ struct ebFrame {
 	jsobjtype winobj;
 	jsobjtype docobj;	/* window.document */
 	const struct MIMETYPE *mt;
+	void *cssmaster;
 };
 extern struct ebFrame *cf;	/* current frame */
 
+#define set_js_globals_f(f) (jcx = f->jcx, winobj = f->winobj, docobj = f->docobj)
+#define set_js_globals() set_js_globals_f(cf)
+#define get_js_globals() (cf->jcx = jcx, cf->winobj = winobj, cf->docobj = docobj)
+
+/* single linked list for internal jump history */
+struct histLabel {
+	int label; /* label must be first element */
+	struct histLabel *prev;
+};
+
 /* an edbrowse window */
 struct ebWindow {
 /* windows stack up as you open new files or follow hyperlinks.
@@ -378,10 +422,9 @@ struct ebWindow {
 	int r_dot, r_dol;
 	struct ebFrame f0; /* first frame */
 	struct ebFrame *jdb_frame; // if in jdb mode
-/* is the referrer the original web page or the individual frame? */
-	char *referrer;
+	char *referrer; // another web page that brought this one to life
 	char *baseDirName;	/* when scanning a directory */
-	char *ft, *fd, *fk;	/* title, description, keywords */
+	char *htmltitle, *htmldesc, *htmlkey;	/* title, description, keywords */
 	char *mailInfo;
 	char lhs[MAXRE], rhs[MAXRE];	/* remembered substitution strings */
 	struct lineMap *map, *r_map;
@@ -393,6 +436,7 @@ struct ebWindow {
 #define MARKLETTERS 27
 #define MARKDOT 26
 	int labels[MARKLETTERS], r_labels[MARKLETTERS];
+	struct histLabel *histLabel;
 /* Next is an array of html tags, generated by the browse command,
  * and used thereafter for hyperlinks, fill-out forms, etc. */
 	struct htmlTag **tags;
@@ -477,7 +521,8 @@ struct htmlTag {
 	struct htmlTag *balance;
 	struct ebFrame *f0; /* frame that owns this tag */
 	struct ebFrame *f1; /* subordinate frame if this is a <frame> tag */
-	jsobjtype jv;		/* corresponding java variable */
+	jsobjtype jv;		/* corresponding javascript variable */
+	jsobjtype style; // style object
 	int seqno;
 	char *js_file;
 	int js_ln;			/* line number of javascript */
@@ -522,28 +567,29 @@ struct htmlTag {
 	uchar itype;		/* input type = */
 	uchar itype_minor;
 	int ninp;		/* number of nonhidden inputs */
-	char *name, *id, *value, *href;
+	char *name, *id, *class, *nodeName, *value, *href;
 	const char *rvalue; /* for reset */
-	char *class;
 	char *innerHTML; /* the html string under this tag */
 	int inner;		/* for inner html */
+	int highspec; // specificity of a selector that matches this node
 };
 
 /* htmlTag.action */
 enum {
-	TAGACT_ZERO, TAGACT_A, TAGACT_INPUT, TAGACT_TITLE, TAGACT_TA,
-	TAGACT_BUTTON, TAGACT_SELECT, TAGACT_OPTION,
+	TAGACT_HTML, TAGACT_A, TAGACT_INPUT, TAGACT_TITLE, TAGACT_TA,
+	TAGACT_BUTTON, TAGACT_SELECT, TAGACT_OPTION, TAGACT_LABEL,
 	TAGACT_NOP, TAGACT_JS, TAGACT_H, TAGACT_SUB, TAGACT_SUP, TAGACT_OVB,
 	TAGACT_OL, TAGACT_UL, TAGACT_DL,
-	TAGACT_TEXT, TAGACT_BODY, TAGACT_HEAD,
+	TAGACT_TEXT, TAGACT_BODY, TAGACT_HEAD, TAGACT_DOC, TAGACT_COMMENT,
 	TAGACT_MUSIC, TAGACT_IMAGE, TAGACT_BR, TAGACT_IBR, TAGACT_P,
-	TAGACT_BASE, TAGACT_META, TAGACT_LINK, TAGACT_PRE, TAGACT_TBODY,
+	TAGACT_BASE, TAGACT_META, TAGACT_LINK, TAGACT_PRE,
+	TAGACT_TBODY, TAGACT_THEAD, TAGACT_TFOOT,
 	TAGACT_DT, TAGACT_DD, TAGACT_LI, TAGACT_TABLE, TAGACT_TR, TAGACT_TD,
 	TAGACT_DIV, TAGACT_SPAN, TAGACT_HR, TAGACT_OBJECT, TAGACT_FOOTER,
 	TAGACT_HEADER, // <header> tag, not the same as <head> tag
 	TAGACT_FORM, TAGACT_FRAME, TAGACT_STYLE,
 	TAGACT_MAP, TAGACT_AREA, TAGACT_SCRIPT, TAGACT_NOSCRIPT, TAGACT_EMBED,
-	TAGACT_OBJ,
+	TAGACT_OBJ, TAGACT_UNKNOWN,
 };
 
 /* htmlTag.itype */
@@ -585,8 +631,4 @@ extern nodeFunction traverse_callback;
 /* Symbolic constants for language independent messages */
 #include "messages.h"
 
-#ifdef __cplusplus
-}
-#endif
-
 #endif
