@@ -53,6 +53,7 @@ static void js_start(void)
 }				/* js_start */
 
 /* Javascript has changed an input field */
+static void javaSetsInner(jsobjtype v, const char *newtext);
 void javaSetsTagVar(jsobjtype v, const char *newtext)
 {
 	struct htmlTag *t = tagFromJavaVar(v);
@@ -62,14 +63,14 @@ void javaSetsTagVar(jsobjtype v, const char *newtext)
 	    || t->itype == INP_FILE)
 		return;
 	if (t->itype == INP_TA) {
-		debugPrint(3, "textarea.value is being updated");
+		javaSetsInner(v, newtext);
 		return;
 	}
 	nzFree(t->value);
 	t->value = cloneString(newtext);
 }				/* javaSetsTagVar */
 
-void javaSetsInner(jsobjtype v, const char *newtext)
+static void javaSetsInner(jsobjtype v, const char *newtext)
 {
 	int side;
 	struct htmlTag *t = tagFromJavaVar(v);
@@ -101,10 +102,6 @@ void dwStart(void)
 	stringAndString(&cf->dw, &cf->dw_l, "<!DOCTYPE public><body>");
 }				/* dwStart */
 
-#define set_js_globals_f(f) (jcx = f->jcx, winobj = f->winobj, docobj = f->docobj)
-#define set_js_globals() set_js_globals_f(cf)
-#define get_js_globals() (cf->jcx = jcx, cf->winobj = winobj, cf->docobj = docobj)
-
 static const char *debugString(const char *v)
 {
 	if (!v)
@@ -145,13 +142,13 @@ q2 while in session 1.
 
 void freeJavaContext(struct ebFrame *f)
 {
-	if (!f->winobj)
+	if (!f->jcx)
 		return;
 	debugPrint(5, "> free frame %p", f);
-	set_js_globals_f(f);
-	freeJavaContext_nat();
+	freeJavaContext_nat(f->jcx);
 	f->jcx = f->winobj = f->docobj = 0;
 	debugPrint(5, "< ok");
+	cssFree(f);
 }				/* freeJavaContext */
 
 /* Run some javascript code under the current window */
@@ -615,35 +612,67 @@ bool run_function_bool(jsobjtype obj, const char *name)
 	return rc;
 }				/* run_function_bool */
 
-bool run_event_bool(jsobjtype obj, const char *pname, const char *evname)
+jsobjtype create_event(jsobjtype parent, const char *evname)
 {
-	bool evdebug = false;
-	if (!handlerPresent(obj, evname))
-		return true;
-	evdebug = get_property_bool(cf->winobj, "eventDebug");
-	if (evdebug)
-		debugPrint(3, "trigger %s.%s", pname, evname);
-	return run_function_bool(obj, evname);
+	jsobjtype e;
+	const char *evname1 = evname;
+	if (evname[0] == 'o' && evname[1] == 'n')
+		evname1 += 2;
+// gc$event protects from garbage collection
+	e = instantiate(parent, "gc$event", "Event");
+	set_property_string(e, "type", evname1);
+	return e;
 }
 
-void run_function_onearg(jsobjtype obj, const char *name, jsobjtype a)
+void unlink_event(jsobjtype parent)
 {
+	delete_property(parent, "gc$event");
+}
+
+bool run_event_bool(jsobjtype obj, const char *pname, const char *evname,
+		    jsobjtype evobj)
+{
+	int rc;
+	jsobjtype eo = evobj;	// event object
+	if (!handlerPresent(obj, evname))
+		return true;
+	if (debugLevel >= 3) {
+		bool evdebug = get_property_bool(cf->winobj, "eventDebug");
+		if (evdebug)
+			debugPrint(3, "trigger %s.%s", pname, evname);
+	}
+// if no event object, create one.
+// It's a one time event that doesn't propagate, like DOMContentLoaded
+// or form.onsubmit etc.
+	if (!eo)
+		eo = create_event(obj, evname);
+	rc = run_function_onearg(obj, evname, eo);
+	if (eo != evobj)
+		unlink_event(obj);
+// no return or some other return is treated as true in this case
+	if (rc < 0)
+		rc = true;
+	return rc;
+}
+
+int run_function_onearg(jsobjtype obj, const char *name, jsobjtype a)
+{
+	int rc;
 	if (!allowJS || !cf->winobj)
-		return;
+		return 0;
 	if (!obj) {
 		debugPrint(3, "run_function_onearg(0, %s", name);
-		return;
-	}
-	if (whichproc == 'j') {
-		run_function_onearg_nat(obj, name, a);
-		return;
+		return 0;
 	}
+	if (whichproc == 'j')
+		return run_function_onearg_nat(obj, name, a);
 	debugPrint(5, "> function %s", name);
 	set_js_globals();
 	whichproc = 'j';	// this line is totally important!
-	run_function_onearg_nat(obj, name, a);
+	rc = run_function_onearg_nat(obj, name, a);
 	whichproc = 'e';
 	debugPrint(5, "< ok");
+	return rc;
 }				/* run_function_onearg */
 
 /*********************************************************************
@@ -663,29 +692,6 @@ void set_basehref(const char *h)
 	set_property_string(cf->winobj, "eb$base", h);
 }				/* set_basehref */
 
-/* The object is a select-one field in the form, and this function returns
- * object.options[selectedIndex].value */
-char *get_property_option(jsobjtype obj)
-{
-	int n;
-	jsobjtype oa;		/* option array */
-	jsobjtype oo;		/* option object */
-
-	if (!allowJS || !cf->winobj || !obj)
-		return 0;
-
-	n = get_property_number(obj, "selectedIndex");
-	if (n < 0)
-		return 0;
-	oa = get_property_object(obj, "options");
-	if (!oa)
-		return 0;
-	oo = get_array_element_object(oa, n);
-	if (!oo)
-		return 0;
-	return get_property_string(oo, "value");
-}				/* get_property_option */
-
 #ifdef DOSLIKE			// port of uname(p), and struct utsname
 struct utsname {
 	char sysname[32];
@@ -717,9 +723,9 @@ void setupJavaDom(void)
 	struct utsname ubuf;
 	int i;
 	char save_c;
-	static const char *languages[] = { 0,
+	static const char *const languages[] = { 0,
 		"english", "french", "portuguese", "polish",
-		"german", "russian",
+		"german", "russian", "italian",
 	};
 	extern const char startWindowJS[];
 	extern const char thirdJS[];
@@ -796,6 +802,8 @@ void setupJavaDom(void)
 		set_property_bool(cf->winobj, "eventDebug", true);
 	if (debugThrow)
 		set_property_bool(cf->winobj, "throwDebug", true);
+	if (uvw)
+		set_property_string(cf->winobj, "$uv$watch", uvw);
 }				/* setupJavaDom */
 
 /* Get the url from a url object, special wrapper.
@@ -874,6 +882,8 @@ static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
 			continue;
 		if (t->controller != sel)
 			continue;
+		if (t->dead)
+			continue;
 
 /* find the corresponding option object */
 		if ((oo = get_array_element_object(oa, i2)) == NULL) {
@@ -885,7 +895,7 @@ static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
 		}
 
 		if (t->jv != oo) {
-// not sure how this would ever happen.
+			debugPrint(5, "oo switch");
 			disconnectTagObject(t);
 			connectTagObject(t, oo);
 		}
@@ -958,13 +968,9 @@ static void rebuildSelector(struct htmlTag *sel, jsobjtype oa, int len2)
 		return;
 	debugPrint(4, "selector %s has changed", selname);
 
-/* If js change the menu, it should have also changed select.value
- * according to the checked options, but did it?
- * Don't know, so I'm going to do it here. */
 	s = displayOptions(sel);
 	if (!s)
 		s = emptyString;
-	set_property_string(sel->jv, "value", s);
 	javaSetsTagVar(sel->jv, s);
 	nzFree(s);
 
