@@ -3,8 +3,9 @@
 /* sourcefile=auth.c */
 bool getUserPass(const char *url, char *creds, bool find_proxy) ;
 bool getUserPassRealm(const char *url, char *creds, const char *realm);
-bool addWebAuthorization(const char *url, const char *credentials, bool proxy,
-			 const char *realm);
+// Add authorization entries only in the foreground, but it's an
+// atomic operation, so the previous routines can run from other threads.
+bool addWebAuthorization(const char *url, const char *credentials, bool proxy, const char *realm);
 
 /* sourcefile=buffers.c */
 void removeHiddenNumbers(pst p, uchar terminate);
@@ -18,7 +19,7 @@ bool cxQuit(int cx, int action) ;
 void cxSwitch(int cx, bool interactive) ;
 bool addTextToBuffer(const pst inbuf, int length, int destl, bool showtrail) ;
 void delText(int start, int end) ;
-bool readFileArgv(const char *filename);
+bool readFileArgv(const char *filename, int fromframe);
 bool unfoldBufferW(const struct ebWindow *w, bool cr, char **data, int *len) ;
 bool unfoldBuffer(int cx, bool cr, char **data, int *len) ;
 bool runCommand(const char *line) ;
@@ -64,7 +65,6 @@ int goSelect(int *startLine, char **rbuf) ;
 
 /* sourcefile=ebjs.c */
 void javaSetsTagVar(jsobjtype v, const char *newtext);
-void javaSetsInner(jsobjtype v, const char *newtext);
 void dwStart(void);
 void createJavaContext(void) ;
 void freeJavaContext(struct ebFrame *f) ;
@@ -92,32 +92,33 @@ jsobjtype instantiate_array_element(jsobjtype array, int idx, const char *classn
 jsobjtype instantiate(jsobjtype parent, const char *name, const char *classname) ;
 int set_property_function(jsobjtype parent, const char *name, const char *body) ;
 int get_arraylength(jsobjtype a);
-char *get_property_option(jsobjtype obj) ;
 void setupJavaDom(void) ;
 char *get_property_url(jsobjtype owner, bool action) ;
 void rebuildSelectors(void);
 bool run_function_bool(jsobjtype obj, const char *name);
-bool run_event_bool(jsobjtype obj, const char *pname, const char *eevname);
-void run_function_onearg(jsobjtype obj, const char *name, jsobjtype o);
+jsobjtype create_event(jsobjtype parent, const char *evname);
+void unlink_event(jsobjtype parent);
+bool run_event_bool(jsobjtype obj, const char *pname, const char *evname, jsobjtype evobj);
+int run_function_onearg(jsobjtype obj, const char *name, jsobjtype o);
 void set_basehref(const char *b);
 
 /* sourcefile=fetchmail.c */
-int fetchMail(int account) ;
-int fetchAllMail(void) ;
-void scanMail(void) ;
-bool emailTest(void) ;
+int fetchMail(int account);
+int fetchAllMail(void);
+void scanMail(void);
+bool emailTest(void);
 void mail64Error(int err);
-char *emailParse(char *buf) ;
-bool setupReply(bool all) ;
+char *emailParse(char *buf);
+bool setupReply(bool all);
 
 /* sourcefile=format.c */
-void prepareForBrowse(char *h, int h_len) ;
+void prepareForBrowse(char *h, int h_len);
 void prepareForField(char *h);
-bool breakLine(const char *line, int len, int *newlen) ;
-void breakLineSetup(void) ;
-char *htmlReformat(char *buf) ;
-void extractEmailAddresses(char *line) ;
-void cutDuplicateEmails(char *tolist, char *cclist, const char *reply) ;
+bool breakLine(const char *line, int len, int *newlen);
+void breakLineSetup(void);
+char *htmlReformat(char *buf);
+void extractEmailAddresses(char *line);
+void cutDuplicateEmails(char *tolist, char *cclist, const char *reply);
 int byteOrderMark(const uchar *buf, int buflen);
 bool looksBinary(const unsigned char *buf, int buflen);
 void looks_8859_utf8(const uchar *buf, int buflen, bool * iso_p, bool * utf8_p);
@@ -184,80 +185,79 @@ void killTag(struct htmlTag *t);
 void html_from_setter( jsobjtype innerParent, const char *h);
 
 /* sourcefile=http.c */
-void gotoLocation(char *url, int delay, bool rf) ;
-size_t eb_curl_callback(char *incoming, size_t size, size_t nitems, struct eb_curl_callback_data *data) ;
+size_t eb_curl_callback(char *incoming, size_t size, size_t nitems, struct i_get *g) ;
 uchar base64Bits(char c);
 int base64Decode(char *start, char **end);
-char *extractHeaderParam(const char *str, const char *item) ;
 time_t parseHeaderDate(const char *date) ;
 bool parseRefresh(char *ref, int *delay_p) ;
-bool shortRefreshDelay(void);
-bool httpConnect(const char *url, bool down_ok, bool webpage, bool f_encoded, char **headers_p, char **body_p, int *bodlen_p);
-void ebcurl_setError(CURLcode curlret, const char *url) ;
-void setHTTPLanguage(const char *lang) ;
+bool shortRefreshDelay(const char *r, int d);
+bool httpConnect(struct i_get *g);
+void ebcurl_setError(CURLcode curlret, const char *url, int action, const char *curl_error);
+void setHTTPLanguage(const char *lang);
 int prompt_and_read(int prompt, char *buffer, int buffer_length, int error_message, bool hide_echo);
-int ebcurl_debug_handler(CURL * handle, curl_infotype info_desc, char *data, size_t size, void *unused) ;
+int ebcurl_debug_handler(CURL * handle, curl_infotype info_desc, char *data, size_t size, struct i_get *g);
 int bg_jobs(bool iponly);
-void addNovsHost(char *host) ;
+void addNovsHost(char *host);
 void deleteNovsHosts(void);
-CURLcode setCurlURL(CURL * h, const char *url) ;
-const char *findProxyForURL(const char *url) ;
+CURLcode setCurlURL(CURL * h, const char *url);
 bool frameExpand(bool expand, int ln1, int ln2);
 int frameExpandLine(int ln, jsobjtype fo);
 struct htmlTag *line2frame(int ln);
 bool reexpandFrame(void);
 bool frameSecurityFile(const char *thisfile);
+void unframe(jsobjtype fobj, jsobjtype newdoc);
+void unframe2(jsobjtype fobj);
 
 /* sourcefile=main.c */
-const char *mailRedirect(const char *to, const char *from, const char *reply, const char *subj) ;
-bool javaOK(const char *url) ;
+const char *mailRedirect(const char *to, const char *from, const char *reply, const char *subj);
+bool javaOK(const char *url);
 void eb_curl_global_init(void);
-void ebClose(int n) ;
-bool isSQL(const char *s) ;
-void setDataSource(char *v) ;
-bool runEbFunction(const char *line) ;
-struct DBTABLE *findTableDescriptor(const char *sn) ;
-struct DBTABLE *newTableDescriptor(const char *name) ;
-bool readConfigFile(void);
+void ebClose(int n);
+bool isSQL(const char *s);
+void setDataSource(char *v);
+bool runEbFunction(const char *line);
+struct DBTABLE *findTableDescriptor(const char *sn);
+struct DBTABLE *newTableDescriptor(const char *name);
+void readConfigFile(void);
 
 /* sourcefile=plugin.c */
-const struct MIMETYPE *findMimeBySuffix(const char *suffix) ;
-const struct MIMETYPE *findMimeByURL(const char *url) ;
-const struct MIMETYPE *findMimeByFile(const char *filename) ;
-const struct MIMETYPE *findMimeByContent(const char *content) ;
-const struct MIMETYPE *findMimeByProtocol(const char *prot) ;
-char *pluginCommand(const struct MIMETYPE *m, const char *infile, const char *outfile, const char *suffix);
+const struct MIMETYPE *findMimeByURL(const char *url, uchar *sxfirst);
+const struct MIMETYPE *findMimeByFile(const char *filename);
+const struct MIMETYPE *findMimeByContent(const char *content);
+bool runPluginCommand(const struct MIMETYPE *m, const char *inurl, const char *infile, const char *indata, int inlength, char **outdata, int *outlength);
 int playBuffer(const char *line, const char *playfile);
-bool playServerData(void);
-char *runPluginConverter(const char *buf, int buflen);
 
 /* sourcefile=sendmail.c */
-bool loadAddressBook(void) ;
-const char *reverseAlias(const char *reply) ;
-bool encodeAttachment(const char *file, int ismail, bool webform, const char **type_p, const char **enc_p, char **data_p) ;
-char *makeBoundary(void) ;
-bool sendMail(int account, const char **recipients, const char *body, int subjat, const char **attachments, const char *refline, int nalt, bool dosig) ;
-bool validAccount(int n) ;
-bool sendMailCurrent(int sm_account, bool dosig) ;
+bool loadAddressBook(void);
+const char *reverseAlias(const char *reply);
+bool encodeAttachment(const char *file, int ismail, bool webform, const char **type_p, const char **enc_p, char **data_p);
+char *makeBoundary(void);
+bool sendMail(int account, const char **recipients, const char *body, int subjat, const char **attachments, const char *refline, int nalt, bool dosig);
+bool validAccount(int n);
+bool sendMailCurrent(int sm_account, bool dosig);
 
 /* sourcefile=messages.c */
-void selectLanguage(void) ;
-const char *i_getString(int msg);
+void eeCheck(void) ;
+void eb_puts(const char *s);
+void selectLanguage(void) ; //@ called once at startup
+const char *i_getString(int msg); //@ the i_ functions are ok in utf8 mode
 void i_puts(int msg) ;
 void i_printf(int msg, ...) ;
 void i_printfExit(int msg, ...) ;
 void i_stringAndMessage(char **s, int *l, int messageNum) ;
-void setError(int msg, ...) ;
+void setError(int msg, ...) ; //? use these only in the foreground!
 void showError(void) ;
 void showErrorConditional(char cmd) ;
 void showErrorAbort(void) ;
 #if 0
 void i_caseShift(unsigned char *s, char action) ;
 #endif
-void eeCheck(void) ;
-void eb_puts(const char *s);
 
-/* sourcefile=stringfile.c */
+// sourcefile=stringfile.c
+// Everything in this file is threadsafe except those marked with @,
+// which are still threadsafe the way we use them, e.g. only calling
+// them from the foreground thread, or ? which is not safe.
+// I'd like to do this markup for the other files as time permits.
 void *allocMem(size_t n) ;
 void *allocZeroMem(size_t n) ;
 void *reallocMem(void *p, size_t n) ;
@@ -309,7 +309,7 @@ bool isA(char c) ;
 bool isquote(char c) ;
 void errorPrint(const char *msg, ...) ;
 void debugPrint(int lev, const char *msg, ...) ;
-void setDebugFile(const char *name);
+void setDebugFile(const char *name); //@
 void nl(void) ;
 int perl2c(char *t) ;
 unsigned pstLength(pst s) ;
@@ -324,24 +324,24 @@ char fileTypeByHandle(int fd) ;
 int fileSizeByName(const char *name) ;
 int fileSizeByHandle(int fd) ;
 time_t fileTimeByName(const char *name) ;
-char *conciseSize(size_t n);
-char *conciseTime(time_t t);
+char *conciseSize(size_t n); //?
+char *conciseTime(time_t t); //?
 bool lsattrChars(const char *buf, char *dest);
-char *lsattr(const char *path, const char *flags);
-void ttySaveSettings(void) ;
+char *lsattr(const char *path, const char *flags); //?
+void ttySaveSettings(void) ; //@
 void ttySetEcho(bool enable_echo);
 #ifndef _INC_CONIO
-int getche(void) ;
-int getch(void) ;
+int getche(void) ; //@
+int getch(void) ; //@
 #endif // #ifndef _INC_CONIO
 char getLetter(const char *s) ;
 char *getFileName(int msg, const char *defname, bool isnew, bool ws);
 int shellProtectLength(const char *s);
 void shellProtect(char *t, const char *s);
-const char *nextScanFile(const char *base);
-bool sortedDirList(const char *dir, struct lineMap **map_p, int *count_p, int othersort, bool reverse) ;
-bool envFile(const char *line, const char **expanded);
-bool envFileDown(const char *line, const char **expanded) ;
+const char *nextScanFile(const char *base); //?
+bool sortedDirList(const char *dir, struct lineMap **map_p, int *count_p, int othersort, bool reverse) ; //?
+bool envFile(const char *line, const char **expanded); //?
+bool envFileDown(const char *line, const char **expanded) ; //?
 FILE *efopen(const char *name, const char *mode) ;
 int eopen(const char *name, int mode, int perms) ;
 void appendFile(const char *fname, const char *message, ...) ;
@@ -359,22 +359,22 @@ char *htmlEscape0(const char *s, bool do_and);
 bool isURL(const char *url) ;
 bool isBrowseableURL(const char *url) ;
 bool isDataURI(const char *u);
-const char *getProtURL(const char *url) ;
-const char *getHostURL(const char *url) ;
-const char *getHostPassURL(const char *url) ;
-const char *getUserURL(const char *url) ;
-const char *getPassURL(const char *url) ;
+const char *getProtURL(const char *url) ; //?
+bool missingProtURL(const char *url);
+const char *getHostURL(const char *url) ; //?
+bool getProtHostURL(const char *url, char *pp, char *hp);
+int getCredsURL(const char *url, char *buf);
 const char *getDataURL(const char *url) ;
 void getDirURL(const char *url, const char **start_p, const char **end_p) ;
 char *findHash(const char *s);
-char *getFileURL(const char *url, bool chophash);
+char *getFileURL(const char *url, bool chophash); //?
 bool getPortLocURL(const char *url, const char **portloc, int *port) ;
 int getPortURL(const char *url) ;
 bool isProxyURL(const char *url) ;
 char *resolveURL(const char *base, const char *rel) ;
 bool sameURL(const char *s, const char *t) ;
-char *altText(const char *base) ;
-char *encodePostData(const char *s) ;
+char *altText(const char *base) ; //?
+char *encodePostData(const char *s, const char *keep_chars) ;
 char *decodePostData(const char *data, const char *name, int seqno) ;
 void decodeMailURL(const char *url, char **addr_p, char **subj_p, char **body_p) ;
 
@@ -384,7 +384,7 @@ void disconnectTagObject(struct htmlTag *t);
 int js_main(void);
 // the native versions of the api functions in ebjs.c
 void createJavaContext_nat(void);
-void freeJavaContext_nat(void);
+void freeJavaContext_nat(jsobjtype cx);
 enum ej_proptype typeof_property_nat(jsobjtype obj, const char *name) ;
 bool has_property_nat(jsobjtype obj, const char *name) ;
 void delete_property_nat(jsobjtype obj, const char *name) ;
@@ -407,6 +407,15 @@ jsobjtype instantiate_nat(jsobjtype parent, const char *name, const char *classn
 int set_property_function_nat(jsobjtype parent, const char *name, const char *body) ;
 int get_arraylength_nat(jsobjtype a);
 bool run_function_bool_nat(jsobjtype obj, const char *name);
-void run_function_onearg_nat(jsobjtype obj, const char *name, jsobjtype o);
+int run_function_onearg_nat(jsobjtype obj, const char *name, jsobjtype o);
 char *run_script_nat(const char *s);
 
+/* sourcefile=css.c */
+void cssDocLoad(jsobjtype thisobj, char *s, bool pageload);
+void cssFree(struct ebFrame *f);
+jsobjtype querySelectorAll(const char *selstring, jsobjtype topobj);
+jsobjtype querySelector(const char *selstring, jsobjtype topobj);
+void cssAttributeCrunch(char *s);
+void cssApply(jsobjtype thisobj, jsobjtype node, jsobjtype destination);
+void cssText(jsobjtype node, const char *rulestring);
+
