@@ -42,10 +42,11 @@ void prepareForBrowse(char *h, int h_len)
 }				/* prepareForBrowse */
 
 /* An input field cannot contain newline, null, or the InternalCodeChar */
+// Revised June 2018, maybe newline is ok. We need it for textarea.
 void prepareForField(char *h)
 {
 	while (*h) {
-		if (*h == 0 || *h == '\n')
+		if (*h == 0)
 			*h = ' ';
 		if (*h == InternalCodeChar)
 			*h = InternalCodeCharAlternate;
@@ -385,6 +386,32 @@ putc:
 		s = w;
 	}
 	*s = 0;
+
+/*********************************************************************
+Some hyperlinks are multiline, due to some html inside, and our interpretation
+of said html. This is just annoying, so pull it back down to one line.
+Same goes for <button>, but other input fields must remain as they are.
+Even submit, as shown by jsrt, if you submit the form it says b1=Send%20Message
+hence it would send a newline if there was one.
+*********************************************************************/
+
+	for (s = buf; (c = *s); ++s) {
+		if (c != InternalCodeChar)
+			continue;
+		n = strtol(s + 1, &s, 10);
+		if (*s == '<') {
+			if (!stringEqual(tagList[n]->info->name, "button"))
+				continue;
+		} else if (*s != '{')
+			continue;
+		for (a = s + 1; (c = *a); ++a) {
+			if (c == InternalCodeChar && a[1] == '0')
+				break;
+			if (c == '\n' || c == '\f')
+				*a = ' ';
+		}
+		s = a;
+	}
 }				/* anchorSwap */
 
 /*********************************************************************
@@ -1021,7 +1048,7 @@ If there is a leading byte order mark as per the previous routine, it's text.
 
 bool looksBinary(const uchar * buf, int buflen)
 {
-	int i, j, bincount = 0, charcount = 0;
+	int i, j, bincount = 0, charcount = 0, nullcount = 0;
 	uchar c;
 	uchar seed;
 
@@ -1031,8 +1058,10 @@ bool looksBinary(const uchar * buf, int buflen)
 	for (i = 0; i < buflen; ++i, ++charcount) {
 		c = buf[i];
 // 0 is ascii, but not really text, and very common in binary files.
-		if (c == 0)
-			++bincount;
+		if (c == 0) {
+			if (++nullcount >= 10)
+				return true;
+		}
 		if (c < 0x80)
 			continue;
 // could represent a utf8 character
@@ -1050,7 +1079,7 @@ binchar:
 		if (seed & 0x80)
 			goto binchar;
 // this is valid utf8 char, don't treat it as binary.
-		i += j;
+		i += j - 1;
 	}
 
 	return (bincount * 8 - 16 >= charcount);
@@ -1505,14 +1534,14 @@ char *force_utf8(char *buf, int buflen)
 		for (s = tbuf; s < tbuf + buflen; ++s)
 			if (!*s)
 				*s = ' ';
-*s = 0;
+		*s = 0;
 		return tbuf;
 	}
 // Strip off the leading bom, if any, and no we're not going to put it back.
 	if (buflen >= 3 && !memcmp(buf, "\xef\xbb\xbf", 3)) {
 		buflen -= 3;
 		memmove(buf, buf + 3, buflen);
-buf[buflen] = 0;
+		buf[buflen] = 0;
 	}
 	for (s = buf; s < buf + buflen; ++s)
 		if (!*s)
@@ -1579,6 +1608,82 @@ char *base64Encode(const char *inbuf, int inlen, bool lines)
 	return outstr;
 }				/* base64Encode */
 
+uchar base64Bits(char c)
+{
+	if (isupperByte(c))
+		return c - 'A';
+	if (islowerByte(c))
+		return c - ('a' - 26);
+	if (isdigitByte(c))
+		return c - ('0' - 52);
+	if (c == '+')
+		return 62;
+	if (c == '/')
+		return 63;
+	return 64;		/* error */
+}				/* base64Bits */
+
+/*********************************************************************
+Decode some data in base64.
+This function operates on the data in-line.  It does not allocate a fresh
+string to hold the decoded data.  Since the data will be smaller than
+the base64 encoded representation, this cannot overflow.
+If you need to preserve the input, copy it first.
+start points to the start of the input
+*end initially points to the byte just after the end of the input
+Returns: GOOD_BASE64_DECODE on success, BAD_BASE64_DECODE or
+EXTRA_CHARS_BASE64_DECODE on error.
+When the function returns success, *end points to the end of the decoded
+data.  On failure, end points to the byte just past the end of
+what was successfully decoded.
+*********************************************************************/
+
+int base64Decode(char *start, char **end)
+{
+	char *b64_end = *end;
+	uchar val, leftover, mod;
+	bool equals;
+	int ret = GOOD_BASE64_DECODE;
+	char c, *q, *r;
+	mod = 0;
+	equals = false;
+	for (q = r = start; q < b64_end; ++q) {
+		c = *q;
+		if (isspaceByte(c))
+			continue;
+		if (equals) {
+			if (c == '=')
+				continue;
+			ret = EXTRA_CHARS_BASE64_DECODE;
+			break;
+		}
+		if (c == '=') {
+			equals = true;
+			continue;
+		}
+		val = base64Bits(c);
+		if (val & 64) {
+			ret = BAD_BASE64_DECODE;
+			break;
+		}
+		if (mod == 0) {
+			leftover = val << 2;
+		} else if (mod == 1) {
+			*r++ = (leftover | (val >> 4));
+			leftover = val << 4;
+		} else if (mod == 2) {
+			*r++ = (leftover | (val >> 2));
+			leftover = val << 6;
+		} else {
+			*r++ = (leftover | val);
+		}
+		++mod;
+		mod &= 3;
+	}
+	*end = r;
+	return ret;
+}				/* base64Decode */
+
 void
 iuReformat(const char *inbuf, int inbuflen, char **outbuf_p, int *outbuflen_p)
 {
