@@ -125,6 +125,7 @@ void html2nodes(const char *htmltext, bool startpage)
 	tidyOptSetBool(tdoc, TidyEscapeScripts, no);
 	tidyOptSetBool(tdoc, TidyDropEmptyElems, no);
 	tidyOptSetBool(tdoc, TidyDropEmptyParas, no);
+	tidyOptSetBool(tdoc, TidyLiteralAttribs, yes);
 
 	tidySetCharEncoding(tdoc, (cons_utf8 ? "utf8" : "latin1"));
 
@@ -266,7 +267,7 @@ static void convertNode(TidyNode node, int level, bool opentag)
 
 	switch (tidyNodeGetType(node)) {
 	case TidyNode_Text:
-		name = "Text";
+		name = "text";
 		break;
 	case TidyNode_Start:
 	case TidyNode_End:
