@@ -26,8 +26,6 @@ Exit codes are as follows:
 
 #include "eb.h"
 
-#include <malloc.h>
-
 #ifdef DOSLIKE
 #include "vsprtf.h"
 #endif // DOSLIKE
@@ -35,6 +33,7 @@ Exit codes are as follows:
 #include <duktape.h>
 
 static void processError(void);
+static void jsInterruptCheck(void);
 
 static duk_ret_t native_error_stub_0(duk_context * cx)
 {
@@ -157,6 +156,38 @@ int js_main(void)
 	return 0;
 }				/* js_main */
 
+// base64 encode
+static duk_ret_t native_btoa(duk_context * cx)
+{
+	char *t;
+	const char *s = duk_get_string(cx, 0);
+	if (!s)
+		s = emptyString;
+	t = base64Encode(s, strlen(s), false);
+	duk_pop(cx);
+	duk_push_string(cx, t);
+	nzFree(t);
+	return 1;
+}
+
+// base64 decode
+static duk_ret_t native_atob(duk_context * cx)
+{
+	char *t1, *t2;
+	const char *s = duk_get_string(cx, 0);
+	if (!s)
+		s = emptyString;
+	t1 = cloneString(s);
+	duk_pop(cx);
+	t2 = t1 + strlen(t1);
+	base64Decode(t1, &t2);
+// ignore errors for now.
+	*t2 = 0;
+	duk_push_string(cx, t1);
+	nzFree(t1);
+	return 1;
+}
+
 static duk_ret_t native_new_location(duk_context * cx)
 {
 	const char *s = duk_safe_to_string(cx, -1);
@@ -186,7 +217,10 @@ static duk_ret_t native_mydoc(duk_context * cx)
 
 static duk_ret_t native_puts(duk_context * cx)
 {
-	printf("%s\n", duk_safe_to_string(cx, -1));
+	const char *s = duk_safe_to_string(cx, -1);
+	if (!s)
+		s = emptyString;
+	puts(s);
 	return 0;
 }
 
@@ -197,6 +231,7 @@ static duk_ret_t native_logputs(duk_context * cx)
 	duk_remove(cx, 0);
 	if (debugLevel >= minlev && s && *s)
 		debugPrint(3, "%s", s);
+	jsInterruptCheck();
 	return 0;
 }
 
@@ -284,9 +319,9 @@ static void jsInterruptCheck(void)
 {
 	if (!intFlag)
 		return;
-	i_puts(MSG_Interrupted);
 	duk_get_global_string(jcx, "eb$stopexec");
 // this next line should fail and stop the script!
+// Assuming we aren't in a try{} block.
 	duk_call(jcx, 0);
 // It didn't stop the script, oh well.
 	duk_pop(jcx);
@@ -313,16 +348,7 @@ static duk_ret_t setter_innerHTML(duk_context * cx)
 	duk_push_this(cx);
 // remove the preexisting children.
 	if (duk_get_prop_string(cx, -1, "childNodes") && duk_is_array(cx, -1)) {
-#if 0
-		int l = duk_get_length(cx, -1);
-// More efficient to remove them last to first.
-		while (l--) {
-			duk_get_prop_index(cx, -1, l);
-			native_removeChild(cx);
-		}
-#else
 		duk_set_length(cx, -1, 0);
-#endif
 	} else {
 // no child nodes array, don't do anything.
 // This should never happen.
@@ -350,31 +376,10 @@ static duk_ret_t setter_innerHTML(duk_context * cx)
 	html_from_setter(thisobj, run);
 	nzFree(run);
 	debugPrint(5, "setter h 2");
-	return 0;
-}
 
-static duk_ret_t getter_innerText(duk_context * cx)
-{
-	duk_push_this(cx);
-	duk_get_prop_string(cx, -1, "inner$Text");
-	duk_remove(cx, -2);
-	return 1;
-}
+	run_function_onearg_nat(context0_obj, "textarea$html$crossover",
+				thisobj);
 
-static duk_ret_t setter_innerText(duk_context * cx)
-{
-	jsobjtype thisobj;
-	const char *h = duk_safe_to_string(cx, -1);
-	if (!h)			// should never happen
-		h = emptyString;
-	debugPrint(5, "setter t 1");
-	duk_push_this(cx);
-	duk_insert(cx, -2);
-	duk_put_prop_string(cx, -2, "inner$Text");
-	thisobj = duk_get_heapptr(cx, -1);
-	duk_pop(cx);
-	javaSetsInner(thisobj, h);
-	debugPrint(5, "setter t 2");
 	return 0;
 }
 
@@ -394,17 +399,16 @@ static duk_ret_t setter_value(duk_context * cx)
 	if (!h)			// should never happen
 		h = emptyString;
 	debugPrint(5, "setter v 1");
+	t = cloneString(h);
 	duk_push_this(cx);
 	duk_insert(cx, -2);
 	duk_put_prop_string(cx, -2, "val$ue");
 	thisobj = duk_get_heapptr(cx, -1);
 	duk_pop(cx);
-	t = cloneString(h);
 	prepareForField(t);
 	debugPrint(4, "value %p=%s", thisobj, t);
 	javaSetsTagVar(thisobj, t);
 	nzFree(t);
-
 	debugPrint(5, "setter v 2");
 	return 0;
 }
@@ -481,6 +485,22 @@ static duk_ret_t setter_cw(duk_context * cx)
 	return 0;
 }
 
+static duk_ret_t native_unframe(duk_context * cx)
+{
+	if (duk_is_object(cx, 0))
+		unframe(duk_get_heapptr(cx, 0), duk_get_heapptr(cx, 1));
+	duk_pop_2(cx);
+	return 0;
+}
+
+static duk_ret_t native_unframe2(duk_context * cx)
+{
+	if (duk_is_object(cx, 0))
+		unframe2(duk_get_heapptr(cx, 0));
+	duk_pop(cx);
+	return 0;
+}
+
 static void linkageNow(char linkmode, jsobjtype o)
 {
 	jsInterruptCheck();
@@ -606,6 +626,9 @@ static void set_timeout(duk_context * cx, bool isInterval)
 		goto done;
 	}
 // stack now has function global fakePropertyName timer-object.
+// classs is milliseconds, for debugging
+	duk_push_int(cx, n);
+	duk_put_prop_string(jcx, -2, "class");
 	to = duk_get_heapptr(cx, -1);
 // protect this timer from the garbage collector.
 	duk_def_prop(cx, 1,
@@ -692,14 +715,18 @@ static void dwrite(duk_context * cx, bool newline)
 		duk_push_string(cx, emptyString);
 	}
 	s = duk_get_string(cx, 0);
+	if (!s || !*s)
+		return;
 	debugPrint(4, "dwrite:%s", s);
 	f = thisFrame(cx);
 	if (!f)
 		debugPrint(3,
 			   "no frame found for document.write, using the default");
 	else {
+#if 0
 		if (f != cf)
 			debugPrint(3, "document.write on a different frame");
+#endif
 		cf = f;
 	}
 	dwStart();
@@ -752,6 +779,7 @@ static const char *embedNodeName(jsobjtype obj)
 		b[length] = 0;
 	}
 	duk_pop_2(jcx);
+	caseShift(b, 'l');
 	return b;
 }				/* embedNodeName */
 
@@ -963,25 +991,25 @@ static duk_ret_t native_removeChild(duk_context * cx)
 	linkageNow('r', thisobj);
 
 done:
-	duk_pop(cx);		// the argument
 	debugPrint(5, "remove 2");
-	return 0;
+	return 1;
 }
 
 static duk_ret_t native_fetchHTTP(duk_context * cx)
 {
 	debugPrint(5, "xhr 1");
 	if (allowXHR) {
+		struct i_get g;
 		const char *incoming_url = duk_safe_to_string(cx, 0);
 		const char *incoming_method = duk_get_string(cx, 1);
 //              const char *incoming_headers = duk_get_string(cx, 2);
 		const char *incoming_payload = duk_get_string(cx, 3);
 		char *outgoing_xhrheaders = NULL;
 		char *outgoing_xhrbody = NULL;
-		int responseLength = 0;
 		char *a = NULL, methchar = '?';
-		bool save_ref, save_plug;
 
+		if (!incoming_url)
+			incoming_url = emptyString;
 		if (incoming_payload && *incoming_payload) {
 			if (incoming_method
 			    && stringEqualCI(incoming_method, "post"))
@@ -993,15 +1021,15 @@ static duk_ret_t native_fetchHTTP(duk_context * cx)
 			incoming_url = a;
 		}
 
-		save_plug = pluginsOn;
-		save_ref = sendReferrer;
 		debugPrint(3, "xhr %s", incoming_url);
-		httpConnect(incoming_url, false, false, true,
-			    &outgoing_xhrheaders, &outgoing_xhrbody,
-			    &responseLength);
+		memset(&g, 0, sizeof(g));
+		g.thisfile = cf->fileName;
+		g.uriEncoded = true;
+		g.url = incoming_url;
+		g.headers_p = &outgoing_xhrheaders;
+		httpConnect(&g);
+		outgoing_xhrbody = g.buffer;
 		nzFree(a);
-		pluginsOn = save_plug;
-		sendReferrer = save_ref;
 		if (intFlag) {
 			duk_get_global_string(cx, "eb$stopexec");
 // this next line should fail and stop the script!
@@ -1035,7 +1063,12 @@ static duk_ret_t native_resolveURL(duk_context * cx)
 {
 	const char *base = duk_get_string(cx, -2);
 	const char *rel = duk_get_string(cx, -1);
-	char *outgoing_url = resolveURL(base, rel);
+	char *outgoing_url;
+	if (!base)
+		base = emptyString;
+	if (!rel)
+		rel = emptyString;
+	outgoing_url = resolveURL(base, rel);
 	if (outgoing_url == NULL)
 		outgoing_url = emptyString;
 	duk_pop_2(cx);
@@ -1079,14 +1112,15 @@ static int cook_l;
 
 static void startCookie(void)
 {
-	nzFree(cookieCopy);
-	cookieCopy = initString(&cook_l);
-	stringAndString(&cookieCopy, &cook_l, "; ");
 	const char *url = cf->fileName;
 	bool secure = false;
 	const char *proto;
 	char *s;
 
+	nzFree(cookieCopy);
+	cookieCopy = initString(&cook_l);
+	stringAndString(&cookieCopy, &cook_l, "; ");
+
 	if (url) {
 		proto = getProtURL(url);
 		if (proto && stringEqualCI(proto, "https"))
@@ -1110,6 +1144,7 @@ static bool foldinCookie(const char *newcook)
 	char *s;
 	char save;
 
+	debugPrint(4, "inject cookie %s", newcook);
 /* make a copy with ; in front */
 	j = strlen(newcook);
 	nc = allocString(j + 3);
@@ -1156,6 +1191,7 @@ add:
 	else
 		stringAndString(&cookieCopy, &cook_l, nc);
 	nzFree(nc);
+	debugPrint(4, "cookieCopy %s", cookieCopy);
 	return true;
 }				/* foldinCookie */
 
@@ -1176,6 +1212,89 @@ static duk_ret_t native_setcook(duk_context * cx)
 	return 0;
 }
 
+static duk_ret_t native_css_start(duk_context * cx)
+{
+	cssDocLoad(duk_get_heapptr(cx, 0), cloneString(duk_get_string(cx, 1)),
+		   duk_get_boolean(cx, 2));
+	return 0;
+}
+
+// querySelectorAll
+static duk_ret_t native_qsa(duk_context * cx)
+{
+	jsobjtype root = 0, ao;
+	const char *selstring = duk_get_string(cx, 0);
+	int top = duk_get_top(cx);
+	if (top > 2) {
+		duk_pop_n(cx, top - 2);
+		top = 2;
+	}
+	if (top == 2) {
+		if (duk_is_object(cx, 1))
+			root = duk_get_heapptr(cx, 1);
+	}
+	if (!root) {
+		duk_push_this(cx);
+		root = duk_get_heapptr(cx, -1);
+		duk_pop(cx);
+	}
+	jsInterruptCheck();
+	ao = querySelectorAll(selstring, root);
+	duk_pop_n(cx, top);
+	duk_push_heapptr(cx, ao);
+	return 1;
+}
+
+// querySelector
+static duk_ret_t native_qs(duk_context * cx)
+{
+	jsobjtype root = 0, ao;
+	const char *selstring = duk_get_string(cx, 0);
+	int top = duk_get_top(cx);
+	if (top > 2) {
+		duk_pop_n(cx, top - 2);
+		top = 2;
+	}
+	if (top == 2) {
+		if (duk_is_object(cx, 1))
+			root = duk_get_heapptr(cx, 1);
+	}
+	if (!root) {
+		duk_push_this(cx);
+		root = duk_get_heapptr(cx, -1);
+		duk_pop(cx);
+	}
+	jsInterruptCheck();
+	ao = querySelector(selstring, root);
+	duk_pop_n(cx, top);
+	if (ao)
+		duk_push_heapptr(cx, ao);
+	else
+		duk_push_undefined(cx);
+	return 1;
+}
+
+static duk_ret_t native_cssApply(duk_context * cx)
+{
+	jsInterruptCheck();
+	if (duk_is_object(cx, 1) && duk_is_object(cx, 2))
+		cssApply(duk_get_heapptr(cx, 0), duk_get_heapptr(cx, 1),
+			 duk_get_heapptr(cx, 2));
+	duk_pop_n(cx, 3);
+	return 0;
+}
+
+static duk_ret_t native_cssText(duk_context * cx)
+{
+	jsobjtype thisobj;
+	const char *rulestring = duk_get_string(cx, 0);
+	duk_push_this(cx);
+	thisobj = duk_get_heapptr(cx, -1);
+	cssText(thisobj, rulestring);
+	duk_pop_2(cx);
+	return 0;
+}
+
 void createJavaContext_nat(void)
 {
 	static int seqno;
@@ -1207,6 +1326,14 @@ void createJavaContext_nat(void)
 	duk_put_global_string(jcx, "my$doc");
 	duk_push_c_function(jcx, native_puts, 1);
 	duk_put_global_string(jcx, "eb$puts");
+	duk_push_c_function(jcx, native_btoa, 1);
+	duk_put_global_string(jcx, "btoa");
+	duk_push_c_function(jcx, native_atob, 1);
+	duk_put_global_string(jcx, "atob");
+	duk_push_c_function(jcx, native_unframe, 2);
+	duk_put_global_string(jcx, "eb$unframe");
+	duk_push_c_function(jcx, native_unframe2, 1);
+	duk_put_global_string(jcx, "eb$unframe2");
 	duk_push_c_function(jcx, native_logputs, 2);
 	duk_put_global_string(jcx, "eb$logputs");
 	duk_push_c_function(jcx, native_prompt, DUK_VARARGS);
@@ -1241,6 +1368,16 @@ void createJavaContext_nat(void)
 	duk_put_global_string(jcx, "eb$getter_cd");
 	duk_push_c_function(jcx, getter_cw, 0);
 	duk_put_global_string(jcx, "eb$getter_cw");
+	duk_push_c_function(jcx, native_css_start, 3);
+	duk_put_global_string(jcx, "eb$cssDocLoad");
+	duk_push_c_function(jcx, native_qsa, DUK_VARARGS);
+	duk_put_global_string(jcx, "querySelectorAll");
+	duk_push_c_function(jcx, native_qs, DUK_VARARGS);
+	duk_put_global_string(jcx, "querySelector");
+	duk_push_c_function(jcx, native_cssApply, 3);
+	duk_put_global_string(jcx, "eb$cssApply");
+	duk_push_c_function(jcx, native_cssText, 1);
+	duk_put_global_string(jcx, "eb$cssText");
 
 	duk_push_heapptr(jcx, docobj);	// native document methods
 	duk_push_c_function(jcx, native_doc_write, DUK_VARARGS);
@@ -1289,11 +1426,11 @@ void createJavaContext_nat(void)
 // setupJavaDom() in ebjs.c does the rest.
 }				/* createJavaContext_nat */
 
-void freeJavaContext_nat(void)
+void freeJavaContext_nat(jsobjtype cx)
 {
 	int i, top = duk_get_top(context0);
 	for (i = 0; i < top; ++i) {
-		if (jcx == duk_get_context(context0, i)) {
+		if (cx == duk_get_context(context0, i)) {
 			duk_remove(context0, i);
 			debugPrint(3, "remove js context %d", i);
 			break;
@@ -1465,17 +1602,20 @@ int set_property_string_nat(jsobjtype parent, const char *name,
 	if (stringEqual(name, "innerHTML"))
 		setter = setter_innerHTML, getter = getter_innerHTML,
 		    altname = "inner$HTML";
-	if (stringEqual(name, "innerText"))
-		setter = setter_innerText, getter = getter_innerText,
-		    altname = "inner$Text";
 	if (stringEqual(name, "value")) {
 // This one is complicated. If option.value had side effects,
 // that would only serve to confuse.
 		bool valsetter = true;
 		duk_get_global_string(jcx, "Option");
-		if (duk_is_array(jcx, -2) || duk_instanceof(jcx, -2, -1))
+		if (duk_instanceof(jcx, -2, -1))
 			valsetter = false;
 		duk_pop(jcx);
+		duk_get_global_string(jcx, "Select");
+		if (duk_instanceof(jcx, -2, -1)) {
+			valsetter = false;
+			puts("select.value set! This shouldn't happen.");
+		}
+		duk_pop(jcx);
 		if (valsetter)
 			setter = setter_value,
 			    getter = getter_value, altname = "val$ue";
@@ -1573,9 +1713,20 @@ int set_property_object_nat(jsobjtype parent, const char *name, jsobjtype child)
 	return 0;
 }				/* set_property_object_nat */
 
+// handler.toString = function() { return this.body; }
+static duk_ret_t native_fntos(duk_context * cx)
+{
+	duk_push_this(cx);
+	duk_get_prop_string(jcx, -1, "body");
+	duk_remove(cx, -2);
+	return 1;
+}
+
 int set_property_function_nat(jsobjtype parent, const char *name,
 			      const char *body)
 {
+	char *body2, *s;
+	int l;
 	if (!body || !*body) {
 // null or empty function, function will return null.
 		body = "null";
@@ -1587,8 +1738,19 @@ int set_property_function_nat(jsobjtype parent, const char *name,
 		debugPrint(3, "compile error for %p.%s", parent, name);
 		duk_push_c_function(jcx, native_error_stub_1, 0);
 	}
-	duk_push_string(jcx, body);
+// At this point I have to undo the mashinations performed by handlerSet().
+	s = body2 = cloneString(body);
+	l = strlen(s);
+	if (l > 16 && stringEqual(s + l - 16, " }.bind(this))()")) {
+		s[l - 16] = 0;
+		if (!strncmp(s, "(function(){", 12))
+			s += 12;
+	}
+	duk_push_string(jcx, s);
+	nzFree(body2);
 	duk_put_prop_string(jcx, -2, "body");
+	duk_push_c_function(jcx, native_fntos, 0);
+	duk_put_prop_string(jcx, -2, "toString");
 	duk_push_heapptr(jcx, parent);
 	duk_insert(jcx, -2);	// switch places
 	duk_put_prop_string(jcx, -2, name);
@@ -1674,10 +1836,43 @@ Function doesn't exist. true, unless debugging.
 Function encounters an error during execution. true, unless debugging.
 *********************************************************************/
 
+// For debugging; please leave the stack the way you found it.
+static void uptrace(jsobjtype node)
+{
+	if (debugLevel < 3)
+		return;
+	duk_push_heapptr(jcx, node);
+	while (true) {
+		const char *nn, *cn;	// node name class name
+		char nnbuf[20];
+		if (duk_get_prop_string(jcx, -1, "nodeName"))
+			nn = duk_to_string(jcx, -1);
+		else
+			nn = "?";
+		strncpy(nnbuf, nn, 20);
+		nnbuf[20 - 1] = 0;
+		if (!nnbuf[0])
+			strcpy(nnbuf, "?");
+		duk_pop(jcx);
+		if (duk_get_prop_string(jcx, -1, "class"))
+			cn = duk_to_string(jcx, -1);
+		else
+			cn = "?";
+		debugPrint(3, "%s.%s", nnbuf, (cn[0] ? cn : "?"));
+		duk_pop(jcx);
+		if (!duk_get_prop_string(jcx, -1, "parentNode")) {
+// we're done.
+			duk_pop_2(jcx);
+			break;
+		}
+		duk_remove(jcx, -2);
+	}
+}
+
 bool run_function_bool_nat(jsobjtype parent, const char *name)
 {
 	int dbl = 3;		// debug level
-	if (stringEqual(name, "ontimer"))
+	if (stringEqual(name, "ontimer") && !uvw)
 		dbl = 4;
 	duk_push_heapptr(jcx, parent);
 	if (!duk_get_prop_string(jcx, -1, name) || !duk_is_function(jcx, -1)) {
@@ -1706,16 +1901,20 @@ bool run_function_bool_nat(jsobjtype parent, const char *name)
 		return rc;
 	}
 // error in execution
+	if (intFlag)
+		i_puts(MSG_Interrupted);
 	processError();
 	debugPrint(3, "failure on %p.%s()", parent, name);
+	uptrace(parent);
 	debugPrint(3, "execution complete");
 	return (debugLevel < 3);
 }				/* run_function_bool_nat */
 
 // The single argument to the function has to be an object.
-void run_function_onearg_nat(jsobjtype parent, const char *name,
-			     jsobjtype child)
+// Returns -1 if the return is not int or bool
+int run_function_onearg_nat(jsobjtype parent, const char *name, jsobjtype child)
 {
+	int rc = -1;
 	duk_push_heapptr(jcx, parent);
 	if (!duk_get_prop_string(jcx, -1, name) || !duk_is_function(jcx, -1)) {
 #if 0
@@ -1723,18 +1922,27 @@ void run_function_onearg_nat(jsobjtype parent, const char *name,
 			asprintf(&errorMessage, "no such function %s", name);
 #endif
 		duk_pop_2(jcx);
-		return;
+		return rc;
 	}
 	duk_insert(jcx, -2);
 	duk_push_heapptr(jcx, child);	// child is the only argument
 	if (!duk_pcall_method(jcx, 1)) {
-// Don't care about the return.
+// See if return is int or bool
+		enum ej_proptype t = top_proptype();
+		if (t == EJ_PROP_BOOL)
+			rc = duk_get_boolean(jcx, -1);
+		if (t == EJ_PROP_INT)
+			rc = duk_get_number(jcx, -1);
 		duk_pop(jcx);
-		return;
+		return rc;
 	}
 // error in execution
+	if (intFlag)
+		i_puts(MSG_Interrupted);
 	processError();
 	debugPrint(3, "failure on %p.%s[]", parent, name);
+	uptrace(parent);
+	return rc;
 }				/* run_function_onearg_nat */
 
 jsobjtype instantiate_array_nat(jsobjtype parent, const char *name)
@@ -1751,6 +1959,7 @@ jsobjtype instantiate_array_nat(jsobjtype parent, const char *name)
 	if (duk_pnew(jcx, 0)) {
 		processError();
 		debugPrint(3, "failure on %p.%s = []", parent, name);
+		uptrace(parent);
 		duk_pop(jcx);
 		return 0;
 	}
@@ -1783,6 +1992,7 @@ jsobjtype instantiate_nat(jsobjtype parent, const char *name,
 		processError();
 		debugPrint(3, "failure on %p.%s = new %s", parent, name,
 			   classname);
+		uptrace(parent);
 		duk_pop(jcx);
 		return 0;
 	}
@@ -1804,6 +2014,7 @@ jsobjtype instantiate_array_element_nat(jsobjtype parent, int idx,
 		processError();
 		debugPrint(3, "failure on %p[%d] = new %s", parent, idx,
 			   classname);
+		uptrace(parent);
 		duk_pop(jcx);
 		return 0;
 	}
@@ -1839,6 +2050,8 @@ char *run_script_nat(const char *s)
 	bool rc;
 	const char *gc;
 	rc = duk_peval_string(jcx, s);
+	if (intFlag)
+		i_puts(MSG_Interrupted);
 	if (!rc) {
 		s = duk_safe_to_string(jcx, -1);
 		if (s && !*s)
