@@ -7,13 +7,14 @@
 
 #include <pthread.h>
 #include <pcre.h>
+#include <signal.h>
 
 /* Define the globals that are declared in eb.h. */
 /* See eb.h for descriptive comments. */
 
 const char *progname;
 const char eol[] = "\r\n";
-const char *version = "3.7.2";
+const char *version = "3.7.4";
 char *changeFileName;
 char *configFile, *addressFile, *cookieFile;
 char *mailDir, *mailUnread, *mailStash, *mailReply;
@@ -22,7 +23,7 @@ char *cacheDir;
 int cacheSize = 1000, cacheCount = 10000;
 char *ebTempDir, *ebUserDir;
 char *userAgents[MAXAGENT + 1];
-char *currentAgent, *currentReferrer;
+char *currentAgent;
 bool allowRedirection = true, allowJS = true, sendReferrer = true;
 bool allowXHR = true;
 bool ftpActive;
@@ -64,7 +65,7 @@ bool inInput, listNA;
 int fileSize;
 char *dbarea, *dblogin, *dbpw;	/* to log into the database */
 bool fetchBlobColumns;
-bool caseInsensitive, searchStringsAll;
+bool caseInsensitive, searchStringsAll, searchWrap = true;
 bool binaryDetect = true;
 bool inputReadLine;
 bool curlAuthNegotiate = false;
@@ -317,13 +318,6 @@ void eb_curl_global_init(void)
 	global_http_handle = curl_easy_init();
 	if (global_http_handle == NULL)
 		goto libcurl_init_fail;
-	if (sslCerts) {
-		curl_init_status =
-		    curl_easy_setopt(global_http_handle, CURLOPT_CAINFO,
-				     sslCerts);
-		if (curl_init_status != CURLE_OK)
-			goto libcurl_init_fail;
-	}
 	if (cookieFile && !ismc) {
 		curl_init_status =
 		    curl_easy_setopt(global_http_handle, CURLOPT_COOKIEJAR,
@@ -430,7 +424,7 @@ static void setupEdbrowseTempDirectory(void)
 int main(int argc, char **argv)
 {
 	int cx, account;
-	bool rc, doConfig = true;
+	bool rc, doConfig = true, autobrowse = false;
 	bool dofetch = false, domail = false;
 	static char agent0[64] = "edbrowse/";
 
@@ -537,14 +531,16 @@ int main(int argc, char **argv)
 	pcre_free = nzFree;
 
 	if (argc && stringEqual(argv[0], "-c")) {
-		if (argc == 1)
+		if (argc == 1) {
 			*argv = configFile;
-		else
-			++argv, --argc;
-		doConfig = false;
-	} else {
-		readConfigFile();
+			doConfig = false;
+		} else {
+			configFile = argv[1];
+			argv += 2, argc -= 2;
+		}
 	}
+	if (doConfig)
+		readConfigFile();
 	account = localAccount;
 
 	for (; argc && argv[0][0] == '-'; ++argv, --argc) {
@@ -571,6 +567,11 @@ int main(int argc, char **argv)
 			continue;
 		}
 
+		if (stringEqual(s, "b")) {
+			autobrowse = true;
+			continue;
+		}
+
 		if (*s == 'p')
 			++s, passMail = true;
 
@@ -605,9 +606,6 @@ int main(int argc, char **argv)
 		i_printfExit(MSG_Usage);
 	}			/* options */
 
-	if (!sslCerts && doConfig && debugLevel >= 1)
-		i_puts(MSG_NoCertFile);
-
 	srand(time(0));
 
 	if (ismc) {
@@ -674,53 +672,82 @@ int main(int argc, char **argv)
 	}
 
 	signal(SIGINT, catchSig);
-#ifndef _MSC_VER		// port siginterrupt(SIGINT, 1); signal(SIGPIPE, SIG_IGN);, if required
-	siginterrupt(SIGINT, 1);
-	signal(SIGPIPE, SIG_IGN);
-#endif // !_MSC_VER
 
 	cx = 0;
 	while (argc) {
 		char *file = *argv;
+		char *file2 = NULL;	// will be allocated
 		++cx;
 		if (cx == MAXSESSION)
 			i_printfExit(MSG_ManyOpen, MAXSESSION);
 		cxSwitch(cx, false);
 		if (cx == 1)
 			runEbFunction("init");
+
+// function on the command line
 		if (file[0] == '<') {
 			runEbFunction(file + 1);
 			++argv, --argc;
 			continue;
 		}
+
 		changeFileName = 0;
-		cf->fileName = cloneString(file);
-		cf->firstURL = cloneString(file);
-		if (isSQL(file))
-			cw->sqlMode = true;
-		rc = readFileArgv(file);
-		if (fileSize >= 0)
-			debugPrint(1, "%d", fileSize);
-		fileSize = -1;
-		if (!rc) {
-			showError();
-		} else if (changeFileName) {
-			nzFree(cf->fileName);
-			cf->fileName = changeFileName;
-			changeFileName = 0;
+		file2 = allocMem(strlen(file) + 10);
+// Every URL needs a protocol.
+		if (missingProtURL(file))
+			sprintf(file2 + 2, "http://%s", file);
+		else
+			strcpy(file2 + 2, file);
+		file = file2 + 2;
+
+		if (autobrowse) {
+			const struct MIMETYPE *mt;
+			uchar sxfirst = 0;
+			if (isURL(file))
+				mt = findMimeByURL(file, &sxfirst);
+			else
+				mt = findMimeByFile(file);
+			if (mt && !mt->outtype)
+				playBuffer("pb", file);
+			else {
+				file2[0] = 'b';
+				file2[1] = ' ';
+				if (runCommand(file2))
+					debugPrint(1, "%d", fileSize);
+				else
+					showError();
+			}
+
+		} else {
+
+			cf->fileName = cloneString(file);
+			cf->firstURL = cloneString(file);
+			if (isSQL(file))
+				cw->sqlMode = true;
+			rc = readFileArgv(file, 0);
+			if (fileSize >= 0)
+				debugPrint(1, "%d", fileSize);
+			fileSize = -1;
+			if (!rc) {
+				showError();
+			} else if (changeFileName) {
+				nzFree(cf->fileName);
+				cf->fileName = changeFileName;
+				changeFileName = 0;
+			}
+			cw->undoable = cw->changeMode = false;
+/* Browse the text if it's a url */
+			if (rc && isURL(cf->fileName)
+			    && ((cf->mt && cf->mt->outtype)
+				|| isBrowseableURL(cf->fileName))) {
+				if (runCommand("b"))
+					debugPrint(1, "%d", fileSize);
+				else
+					showError();
+			}
 		}
 
-		cw->undoable = cw->changeMode = false;
-/* Browse the text if it's a url */
-		if (rc && isURL(cf->fileName) && ((cf->mt && cf->mt->outtype)
-						  ||
-						  isBrowseableURL
-						  (cf->fileName))) {
-			if (runCommand("b"))
-				debugPrint(1, "%d", fileSize);
-			else
-				showError();
-		}
+		nzFree(file2);
 		++argv, --argc;
 	}			/* loop over files */
 	if (!cx) {		/* no files */
@@ -1061,13 +1088,13 @@ static const char *const keywords[] = {
 
 /* Read the config file and populate the corresponding data structures. */
 /* This routine succeeds, or aborts via one of these macros. */
-#define cfgAbort0(m) { i_printf(m); nl(); return false; }
-#define cfgAbort1(m, arg) { i_printf(m, arg); nl(); return false; }
-#define cfgLine0(m) { i_printf(m, ln); nl(); return false; }
-#define cfgLine1(m, arg) { i_printf(m, ln, arg); nl(); return false; }
-#define cfgLine1a(m, arg) { i_printf(m, arg, ln); nl(); return false; }
+#define cfgAbort0(m) { i_printf(m); nl(); return; }
+#define cfgAbort1(m, arg) { i_printf(m, arg); nl(); return; }
+#define cfgLine0(m) { i_printf(m, ln); nl(); return; }
+#define cfgLine1(m, arg) { i_printf(m, ln, arg); nl(); return; }
+#define cfgLine1a(m, arg) { i_printf(m, arg, ln); nl(); return; }
 
-bool readConfigFile(void)
+void readConfigFile(void)
 {
 	char *buf, *s, *t, *v, *q;
 	int buflen, n;
@@ -1088,10 +1115,11 @@ bool readConfigFile(void)
 
 	unreadConfigFile();
 
-	if (!fileTypeByName(configFile, false))
-		return true;	/* config file not present */
-	if (!fileIntoMemory(configFile, &buf, &buflen))
-		showErrorAbort();
+	if (!fileIntoMemory(configFile, &buf, &buflen)) {
+		i_printf(MSG_NoConfig, configFile);
+		return;
+	}
+
 /* An extra newline won't hurt. */
 	if (buflen && buf[buflen - 1] != '\n')
 		buf[buflen++] = '\n';
@@ -1398,7 +1426,6 @@ putc:
 
 		case 11:	/* protocol */
 			mt->prot = v;
-			mt->stream = true;
 			continue;
 
 		case 12:	/* program */
@@ -1622,12 +1649,12 @@ nokeyword:
 			continue;
 		}
 
-		if (stringEqual(s, "download") && mimeblock == 1) {
-			mt->download = true;
+		if (stringEqual(s, "from_file") && mimeblock == 1) {
+			mt->from_file = true;
 			continue;
 		}
-		if (stringEqual(s, "stream") && mimeblock == 1) {
-			mt->stream = true;
+		if (stringEqual(s, "down_url") && mimeblock == 1) {
+			mt->down_url = true;
 			continue;
 		}
 
@@ -1797,10 +1824,6 @@ putback:
 	if (mailblock | mimeblock)
 		cfgAbort0(MSG_EBRC_MNotClosed);
 
-	if (!sslCerts)
-		verifyCertificates = 0;
-
 	if (maxAccount && !localAccount)
 		localAccount = 1;
-	return true;
 }				/* readConfigFile */
