@@ -116,11 +116,28 @@ void selectLanguage(void)
 		return;
 	}
 
+	if (!strncmp(eb_language, "it", 2)) {
+		eb_lang = 7;
+		messageArray = msg_it;
+		type8859 = 1;
+		return;
+	}
+
 /* This error is really annoying if it pops up every time you invoke edbrowse.
 	fprintf(stderr, "Sorry, language %s is not implemented\n", buf);
 */
 }				/* selectLanguage */
 
+/*********************************************************************
+WARNING: this routine, which is at the heart of the international prints
+i_puts i_printf, is not threadsafe in iso8859 mode.
+Well utf8 has been the default console standard for 15 years now,
+and I'm almost ready to chuck iso8859 altogether, so for now,
+let's just say you can't use threading in 8859 mode.
+If you try to turn it on via the bg (background) command, I won't let you.
+I really don't think this will come up, everybody is utf8 by now.
+*********************************************************************/
+
 const char *i_getString(int msg)
 {
 	const char **a = messageArray;
@@ -210,7 +227,7 @@ Thus I don't need the i_ prefix.
 char errorMsg[1024];
 
 /* Show the error message, not just the question mark, after these commands. */
-static const char showerror_cmd[] = "AbefMqrw^";
+static const char showerror_cmd[] = "AbefMqrw^&";
 
 /* Set the error message.  Type h to see the message. */
 void setError(int msg, ...)
