@@ -15,12 +15,15 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <malloc.h>
-
+#ifdef _MSC_VER
+#include <WinSock2.h>
+#include <sys/timeb.h>
+#endif /* _MSC_VER y/n */
 #include "vsprtf.h"
 
 static const char *module = "vsprtf";
 
-#ifdef _MSC_VER
+#if (defined(_MSC_VER) && (_MSC_VER < 1900))
 #define va_copy(dst,src) dst = src
 #endif
 
@@ -74,4 +77,28 @@ int vasprintf (char **str, const char *fmt, va_list args)
     return size;
 }
 
+/* SPECIAL CASE ONLY */
+#if (defined(_MSC_VER) && defined(NDEBUG))
+/*  ***********************************************************************
+    20171108: NOTE: At present this is defined in static libtidy, but only 
+    in the Debug version. So this is a substitute for other than
+    the Debug version. This is an 'unofficial' extern in static Debug 
+    libtidy, NOT part of the API, so may change in future!
+    *********************************************************************** */
+int gettimeofday(struct timeval *tp, void *tzp)
+{
+#ifdef WIN32
+    struct _timeb timebuffer;
+    _ftime(&timebuffer);
+    tp->tv_sec = (long)timebuffer.time;
+    tp->tv_usec = timebuffer.millitm * 1000;
+#else
+    tp->tv_sec = time(NULL);
+    tp->tv_usec = 0;
+#endif
+    return 0;
+}
+
+#endif /* _MSC_VER */
+
 // eof = vsprtf.cxx
